# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import errno
import filecmp
import os
import pathlib
import socket
import subprocess
import sys
from collections.abc import Callable
from unittest.mock import call, patch

import pytest

from charmcraft import charm_builder, const
from charmcraft.charm_builder import (
    KNOWN_GOOD_PIP_URL,
    CharmBuilder,
    _process_run,
)


def test_build_generics_simple_files(tmp_path):
    """Check transferred metadata and simple entrypoint, also return proper linked entrypoint."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")

    entrypoint = tmp_path / "crazycharm.py"
    entrypoint.touch()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=entrypoint,
    )
    linked_entrypoint = builder.handle_generic_paths()

    # check files are there, are files, and are really hard links (so no
    # check for permissions needed)
    built_metadata = build_dir / const.METADATA_FILENAME
    assert built_metadata.is_file()
    assert built_metadata.stat().st_ino == metadata.stat().st_ino

    built_entrypoint = build_dir / "crazycharm.py"
    assert built_entrypoint.is_file()
    assert built_entrypoint.stat().st_ino == entrypoint.stat().st_ino

    assert linked_entrypoint == built_entrypoint


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_simple_dir(tmp_path):
    """Check transferred any directory, with proper permissions."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()
    entrypoint = tmp_path / "crazycharm.py"
    entrypoint.touch()
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")

    somedir = tmp_path / "somedir"
    somedir.mkdir(mode=0o700)

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=entrypoint,
    )
    builder.handle_generic_paths()

    built_dir = build_dir / "somedir"
    assert built_dir.is_dir()
    assert built_dir.stat().st_mode & 0xFFF == 0o700


def test_build_generics_ignored_file(tmp_path, assert_output):
    """Don't include ignored filed."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")

    # create two files (and the needed entrypoint)
    file1 = tmp_path / "file1.txt"
    file1.touch()
    file2 = tmp_path / "file2.txt"
    file2.touch()
    entrypoint = tmp_path / "crazycharm.py"
    entrypoint.touch()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=entrypoint,
    )

    # set it up to ignore file 2 and make it work
    builder.ignore_rules.extend_patterns(["file2.*"])
    builder.handle_generic_paths()

    assert (build_dir / "file1.txt").exists()
    assert not (build_dir / "file2.txt").exists()

    expected = "Ignoring file because of rules: 'file2.txt'"
    assert_output(expected)


def test_build_generics_ignored_dir(tmp_path, assert_output):
    """Don't include ignored dir."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")

    # create two files (and the needed entrypoint)
    dir1 = tmp_path / "dir1"
    dir1.mkdir()
    dir2 = tmp_path / "dir2"
    dir2.mkdir()
    entrypoint = tmp_path / "crazycharm.py"
    entrypoint.touch()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=entrypoint,
    )

    # set it up to ignore dir 2 and make it work
    builder.ignore_rules.extend_patterns(["dir2"])
    builder.handle_generic_paths()

    assert (build_dir / "dir1").exists()
    assert not (build_dir / "dir2").exists()

    expected = "Ignoring directory because of rules: 'dir2'"
    assert_output(expected)


def _test_build_generics_tree(tmp_path, *, expect_hardlinks):
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    # create this structure:
    # ├─ crazycharm.py  (entrypoint)
    # ├─ file1.txt
    # ├─ dir1
    # │  └─ dir3  (ignored!)
    # └─ dir2
    #    ├─ file2.txt
    #    ├─ file3.txt  (ignored!)
    #    ├─ dir4  (ignored!)
    #    │   └─ file4.txt
    #    └─ dir5
    entrypoint = tmp_path / "crazycharm.py"
    entrypoint.touch()
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    file1 = tmp_path / "file1.txt"
    file1.touch()
    dir1 = tmp_path / "dir1"
    dir1.mkdir()
    dir3 = dir1 / "dir3"
    dir3.mkdir()
    dir2 = tmp_path / "dir2"
    dir2.mkdir()
    file2 = dir2 / "file2.txt"
    file2.touch()
    file3 = dir2 / "file3.txt"
    file3.touch()
    dir4 = dir2 / "dir4"
    dir4.mkdir()
    file4 = dir4 / "file4.txt"
    file4.touch()
    dir5 = dir2 / "dir5"
    dir5.mkdir()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=entrypoint,
    )

    # set it up to ignore some stuff and make it work
    builder.ignore_rules.extend_patterns(
        [
            "dir1/dir3",
            "dir2/file3.txt",
            "dir2/dir4",
        ]
    )
    builder.handle_generic_paths()

    assert (build_dir / "crazycharm.py").exists()
    assert (build_dir / "file1.txt").exists()
    assert (build_dir / "dir1").exists()
    assert not (build_dir / "dir1" / "dir3").exists()
    assert (build_dir / "dir2").exists()
    assert (build_dir / "dir2" / "file2.txt").exists()
    assert not (build_dir / "dir2" / "file3.txt").exists()
    assert not (build_dir / "dir2" / "dir4").exists()
    assert (build_dir / "dir2" / "dir5").exists()

    for p1, p2 in [
        (build_dir / "crazycharm.py", entrypoint),
        (build_dir / "file1.txt", file1),
        (build_dir / "dir2" / "file2.txt", file2),
    ]:
        if expect_hardlinks:
            # they're hard links
            assert p1.samefile(p2)
        else:
            # they're *not* hard links
            assert not p1.samefile(p2)
            # but they're essentially the same
            assert filecmp.cmp(str(p1), str(p2), shallow=False)
            assert p1.stat().st_mode == p2.stat().st_mode
            assert p1.stat().st_size == p2.stat().st_size
            assert p1.stat().st_atime == pytest.approx(p2.stat().st_atime)
            assert p1.stat().st_mtime == pytest.approx(p2.stat().st_mtime)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_tree(tmp_path):
    """Manages ok a deep tree, including internal ignores."""
    _test_build_generics_tree(tmp_path, expect_hardlinks=True)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_tree_vagrant(tmp_path):
    """Manages ok a deep tree, including internal ignores, when hardlinks aren't allowed."""
    with patch("os.link") as mock_link:
        mock_link.side_effect = PermissionError("No you don't.")
        _test_build_generics_tree(tmp_path, expect_hardlinks=False)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_tree_xdev(tmp_path):
    """Manages ok a deep tree, including internal ignores, when hardlinks can't be done."""
    with patch("os.link") as mock_link:
        mock_link.side_effect = OSError(errno.EXDEV, os.strerror(errno.EXDEV))
        _test_build_generics_tree(tmp_path, expect_hardlinks=False)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_symlink_file(tmp_path):
    """Respects a symlinked file."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    entrypoint = tmp_path / "crazycharm.py"
    entrypoint.touch()
    the_symlink = tmp_path / "somehook.py"
    the_symlink.symlink_to(entrypoint)

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=entrypoint,
    )
    builder.handle_generic_paths()

    built_symlink = build_dir / "somehook.py"
    assert built_symlink.is_symlink()
    assert built_symlink.resolve() == build_dir / "crazycharm.py"
    assert built_symlink.readlink() == pathlib.Path("crazycharm.py")


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_symlink_dir(tmp_path):
    """Respects a symlinked dir."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    entrypoint = tmp_path / "crazycharm.py"
    entrypoint.touch()
    somedir = tmp_path / "somedir"
    somedir.mkdir()
    somefile = somedir / "some file"
    somefile.touch()
    the_symlink = tmp_path / "thelink"
    the_symlink.symlink_to(somedir)

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=entrypoint,
    )
    builder.handle_generic_paths()

    built_symlink = build_dir / "thelink"
    assert built_symlink.is_symlink()
    assert built_symlink.resolve() == build_dir / "somedir"
    assert built_symlink.readlink() == pathlib.Path("somedir")

    # the file inside the linked dir should exist
    assert (build_dir / "thelink" / "some file").exists()


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_symlink_deep(tmp_path):
    """Correctly re-links a symlink across deep dirs."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    entrypoint = tmp_path / "crazycharm.py"
    entrypoint.touch()

    dir1 = tmp_path / "dir1"
    dir1.mkdir()
    dir2 = tmp_path / "dir2"
    dir2.mkdir()
    original_target = dir1 / "file.real"
    original_target.touch()
    the_symlink = dir2 / "file.link"
    the_symlink.symlink_to(original_target)

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=entrypoint,
    )
    builder.handle_generic_paths()

    built_symlink = build_dir / "dir2" / "file.link"
    assert built_symlink.is_symlink()
    assert built_symlink.resolve() == build_dir / "dir1" / "file.real"
    assert built_symlink.readlink() == pathlib.Path("..", "dir1", "file.real")


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_symlink_file_outside(tmp_path, assert_output):
    """Ignores (with warning) a symlink pointing a file outside projects dir."""
    project_dir = tmp_path / "test-project"
    project_dir.mkdir()

    metadata = project_dir / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = project_dir / const.BUILD_DIRNAME
    build_dir.mkdir()
    entrypoint = project_dir / "crazycharm.py"
    entrypoint.touch()

    outside_project = tmp_path / "dangerous.txt"
    outside_project.touch()
    the_symlink = project_dir / "external-file"
    the_symlink.symlink_to(outside_project)

    builder = CharmBuilder(
        builddir=project_dir,
        installdir=build_dir,
        entrypoint=entrypoint,
    )
    builder.handle_generic_paths()

    assert not (build_dir / "external-file").exists()
    expected = "Ignoring symlink because targets outside the project: 'external-file'"
    assert_output(expected)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_symlink_directory_outside(tmp_path, assert_output):
    """Ignores (with warning) a symlink pointing a dir outside projects dir."""
    project_dir = tmp_path / "test-project"
    project_dir.mkdir()

    metadata = project_dir / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = project_dir / const.BUILD_DIRNAME
    build_dir.mkdir()
    entrypoint = project_dir / "crazycharm.py"
    entrypoint.touch()

    outside_project = tmp_path / "dangerous"
    outside_project.mkdir()
    the_symlink = project_dir / "external-dir"
    the_symlink.symlink_to(outside_project)

    builder = CharmBuilder(
        builddir=project_dir,
        installdir=build_dir,
        entrypoint=entrypoint,
    )
    builder.handle_generic_paths()

    assert not (build_dir / "external-dir").exists()
    expected = "Ignoring symlink because targets outside the project: 'external-dir'"
    assert_output(expected)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_generics_different_filetype(tmp_path, assert_output, monkeypatch):
    """Ignores whatever is not a regular file, symlink or dir."""
    # change into the tmp path and do everything locally, because otherwise the socket path
    # will be too long for mac os
    monkeypatch.chdir(tmp_path)

    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = pathlib.Path(const.BUILD_DIRNAME)
    build_dir.mkdir()
    entrypoint = pathlib.Path("crazycharm.py")
    entrypoint.touch()

    # create a socket
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind("test-socket")

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=tmp_path / entrypoint,
    )
    builder.handle_generic_paths()

    assert not (build_dir / "test-socket").exists()
    expected = "Ignoring file because of type: 'test-socket'"
    assert_output(expected)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_dispatcher_modern_dispatch_created(tmp_path):
    """The dispatcher script is properly built."""
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    linked_entrypoint = build_dir / "somestuff.py"

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
    )
    builder.handle_dispatcher(linked_entrypoint)

    included_dispatcher = build_dir / const.DISPATCH_FILENAME
    with included_dispatcher.open("rt", encoding="utf8") as fh:
        dispatcher_code = fh.read()
    assert dispatcher_code == const.DISPATCH_CONTENT.format(
        entrypoint_relative_path="somestuff.py"
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_dispatcher_modern_dispatch_respected(tmp_path):
    """The already included dispatcher script is left untouched."""
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    already_present_dispatch = build_dir / const.DISPATCH_FILENAME
    with already_present_dispatch.open("wb") as fh:
        fh.write(b"abc")

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
    )
    builder.handle_dispatcher("whatever")

    with already_present_dispatch.open("rb") as fh:
        assert fh.read() == b"abc"


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_dispatcher_classic_hooks_mandatory_created(tmp_path):
    """The mandatory classic hooks are implemented ok if not present."""
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    linked_entrypoint = build_dir / "somestuff.py"
    included_dispatcher = build_dir / const.DISPATCH_FILENAME

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
    )
    with patch("charmcraft.const.MANDATORY_HOOK_NAMES", {"testhook"}):
        builder.handle_dispatcher(linked_entrypoint)

    test_hook = build_dir / const.HOOKS_DIRNAME / "testhook"
    assert test_hook.is_symlink()
    assert test_hook.resolve() == included_dispatcher
    assert test_hook.readlink() == pathlib.Path("..", const.DISPATCH_FILENAME)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_dispatcher_classic_hooks_mandatory_respected(tmp_path):
    """The already included mandatory classic hooks are left untouched."""
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    built_hooks_dir = build_dir / const.HOOKS_DIRNAME
    built_hooks_dir.mkdir()
    test_hook = built_hooks_dir / "testhook"
    with test_hook.open("wb") as fh:
        fh.write(b"abc")

    linked_entrypoint = build_dir / "somestuff.py"

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
    )
    with patch("charmcraft.const.MANDATORY_HOOK_NAMES", {"testhook"}):
        builder.handle_dispatcher(linked_entrypoint)

    with test_hook.open("rb") as fh:
        assert fh.read() == b"abc"


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_dispatcher_classic_hooks_linking_charm_replaced(tmp_path, assert_output):
    """Hooks that are just a symlink to the entrypoint are replaced."""
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    # simple source code
    src_dir = build_dir / "src"
    src_dir.mkdir()
    built_charm_script = src_dir / "charm.py"
    with built_charm_script.open("wb") as fh:
        fh.write(b"all the magic")

    # a test hook, just a symlink to the charm
    built_hooks_dir = build_dir / const.HOOKS_DIRNAME
    built_hooks_dir.mkdir()
    test_hook = built_hooks_dir / "somehook"
    test_hook.symlink_to(built_charm_script)

    included_dispatcher = build_dir / const.DISPATCH_FILENAME

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
    )
    builder.handle_dispatcher(built_charm_script)

    # the test hook is still there and a symlink, but now pointing to the dispatcher
    assert test_hook.is_symlink()
    assert test_hook.resolve() == included_dispatcher
    expected = "Replacing existing hook 'somehook' as it's a symlink to the entrypoint"
    assert_output(expected)


# -- tests about dependencies handling


@pytest.mark.parametrize(
    ("python_packages", "binary_packages", "reqs_contents", "charmlibs", "expected_call_params"),
    [
        pytest.param(
            [],
            [],
            [],
            [],
            [["install", "--no-binary=:all:", "--requirement={reqs_file}"]],
            id="simple",
        ),
        pytest.param(
            ["pkg1", "pkg2"],
            [],
            [],
            [],
            [
                ["install", "--no-binary=:all:", "pkg1", "pkg2"],
                ["install", "--no-binary=:all:", "--requirement={reqs_file}"],
            ],
            id="packages-only",
        ),
        pytest.param(
            [],
            ["bin-pkg1", "bin-pkg2"],
            [],
            [],
            [
                ["install", "bin-pkg1", "bin-pkg2"],
                ["install", "--no-binary=:all:", "--requirement={reqs_file}"],
            ],
            id="binary-packages-only",
        ),
        pytest.param(
            ["pkg1", "pkg2"],
            ["bin-pkg1", "bin-pkg2"],
            [],
            [],
            [
                ["install", "bin-pkg1", "bin-pkg2"],
                ["install", "--no-binary=:all:", "pkg1", "pkg2"],
                ["install", "--no-binary=:all:", "--requirement={reqs_file}"],
            ],
            id="binary-and-source-packages",
        ),
        pytest.param(
            [],
            [],
            ["req1", "req2"],
            [],
            [
                ["install", "--no-binary=:all:", "--requirement={reqs_file}"],
            ],
            id="requirements-only",
        ),
        pytest.param(
            ["req1"],
            ["req2"],
            ["req1", "req2"],
            [],
            [
                ["install", "req2"],
                ["install", "--no-binary=:all:", "req1"],
                ["install", "--no-binary=:all:", "--requirement={reqs_file}"],
            ],
            id="requirements-duplicated-in-charmcraft_yaml",
        ),
        pytest.param(
            [],
            [],
            [],
            ["charmlib-dep"],
            [["install", "--no-binary=:all:", "--requirement={reqs_file}", "charmlib-dep"]],
            id="charmlib-dep-only",
        ),
        pytest.param(
            [],
            [],
            ["charmlib-dep==0.1", "req1"],
            ["charmlib-dep"],
            [["install", "--no-binary=:all:", "--requirement={reqs_file}"]],
            id="charmlib-dep-in-requirements",
        ),
        pytest.param(
            ["duplicate"],
            ["duplicate"],
            ["duplicate"],
            ["duplicate"],
            [
                ["install", "duplicate"],
                ["install", "--no-binary=:all:", "duplicate"],
                ["install", "--no-binary=:all:", "--requirement={reqs_file}"],
            ],
            id="all-same",
        ),
        pytest.param(
            ["duplicate", "pkg1"],
            ["duplicate", "bin-pkg1"],
            ["duplicate", "req1"],
            ["duplicate", "lib-dep"],
            [
                ["install", "bin-pkg1", "duplicate"],
                ["install", "--no-binary=:all:", "duplicate", "pkg1"],
                ["install", "--no-binary=:all:", "--requirement={reqs_file}", "lib-dep"],
            ],
            id="all-overlap",
        ),
    ],
)
def test_build_dependencies_virtualenv(
    tmp_path: pathlib.Path,
    assert_output: Callable,
    python_packages: list[str],
    binary_packages: list[str],
    reqs_contents: list[str],
    charmlibs: list[str],
    expected_call_params,
):
    """A virtualenv is created with the specified requirements file."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    reqs_file = tmp_path / "reqs.txt"
    reqs_file.write_text("\n".join(reqs_contents))

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
        binary_python_packages=binary_packages,
        python_packages=python_packages,
        requirements=[reqs_file],
    )
    builder.charmlib_deps = set(charmlibs)

    with patch("charmcraft.charm_builder.get_pip_version") as mock_pip_version:
        mock_pip_version.return_value = (22, 0)
        with patch("charmcraft.charm_builder._process_run") as mock:
            with patch("shutil.copytree") as mock_copytree:
                builder.handle_dependencies()

    pip_cmd = str(charm_builder._find_venv_bin(tmp_path / const.STAGING_VENV_DIRNAME, "pip"))

    formatted_calls = [
        [param.format(reqs_file=str(reqs_file)) for param in call] for call in expected_call_params
    ]
    extra_pip_calls = [call([pip_cmd, *params]) for params in formatted_calls]

    assert mock.mock_calls == [
        call(["python3", "-m", "venv", str(tmp_path / const.STAGING_VENV_DIRNAME)]),
        call([pip_cmd, "install", f"pip@{KNOWN_GOOD_PIP_URL}"]),
        *extra_pip_calls,
    ]

    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]
    assert_output("Handling dependencies", "Installing dependencies")


def test_build_dependencies_virtualenv_multiple(tmp_path, assert_output):
    """A virtualenv is created with multiple requirements files."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    reqs_file_1 = tmp_path / "reqs.txt"
    reqs_file_1.touch()
    reqs_file_2 = tmp_path / "reqs.txt"
    reqs_file_1.touch()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
        binary_python_packages=[],
        python_packages=[],
        requirements=[reqs_file_1, reqs_file_2],
    )

    with patch("charmcraft.charm_builder.get_pip_version") as mock_pip_version:
        mock_pip_version.return_value = (22, 0)
        with patch("charmcraft.charm_builder._process_run") as mock:
            with patch("shutil.copytree") as mock_copytree:
                builder.handle_dependencies()

    pip_cmd = str(charm_builder._find_venv_bin(tmp_path / const.STAGING_VENV_DIRNAME, "pip"))
    assert mock.mock_calls == [
        call(["python3", "-m", "venv", str(tmp_path / const.STAGING_VENV_DIRNAME)]),
        call([pip_cmd, "install", f"pip@{KNOWN_GOOD_PIP_URL}"]),
        call(
            [
                pip_cmd,
                "install",
                "--no-binary=:all:",
                f"--requirement={reqs_file_1}",
                f"--requirement={reqs_file_2}",
            ]
        ),
    ]

    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]
    assert_output("Handling dependencies", "Installing dependencies")


def test_build_dependencies_virtualenv_none(tmp_path, assert_output):
    """The virtualenv is NOT created if no needed."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
        binary_python_packages=[],
        python_packages=[],
        requirements=[],
    )

    with patch("charmcraft.charm_builder.subprocess.run") as mock_run:
        builder.handle_dependencies()

    mock_run.assert_not_called()
    assert_output("Handling dependencies", "No dependencies to handle")


def test_build_dependencies_no_reused_missing_venv(tmp_path, assert_output):
    """Dependencies are built again because installation dir was not found."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
        binary_python_packages=[],
        python_packages=["ops"],
        requirements=[],
    )
    staging_venv_dir = tmp_path / const.STAGING_VENV_DIRNAME

    # patch the dependencies installation method so it skips all subprocessing but actually
    # creates the directory, to simplify testing
    builder._install_dependencies = lambda dirpath: dirpath.mkdir()

    # first run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Dependencies directory not found",
        "Installing dependencies",
    )

    # directory created and packages installed
    assert staging_venv_dir.exists()

    # installation directory copied to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]

    # remove the site venv directory
    staging_venv_dir.rmdir()

    # second run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Dependencies directory not found",
        "Installing dependencies",
    )

    # directory created and packages installed *again*
    assert staging_venv_dir.exists()

    # installation directory copied *again* to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]


def test_build_dependencies_no_reused_missing_hash_file(tmp_path, assert_output):
    """Dependencies are built again because previous hash file was not found."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
        binary_python_packages=[],
        python_packages=["ops"],
        requirements=[],
    )
    staging_venv_dir = tmp_path / const.STAGING_VENV_DIRNAME

    # patch the dependencies installation method so it skips all subprocessing but actually
    # creates the directory, to simplify testing
    builder._install_dependencies = lambda dirpath: dirpath.mkdir(exist_ok=True)

    # first run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Dependencies directory not found",
        "Installing dependencies",
    )

    # directory created and packages installed
    assert staging_venv_dir.exists()

    # installation directory copied to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]

    # remove the hash file
    (tmp_path / const.DEPENDENCIES_HASH_FILENAME).unlink()

    # second run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Dependencies hash file not found",
        "Installing dependencies",
    )

    # directory created and packages installed *again*
    assert staging_venv_dir.exists()

    # installation directory copied *again* to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]


def test_build_dependencies_no_reused_problematic_hash_file(tmp_path, assert_output):
    """Dependencies are built again because having problems to read the previous hash file."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
        binary_python_packages=[],
        python_packages=["ops"],
        requirements=[],
    )
    staging_venv_dir = tmp_path / const.STAGING_VENV_DIRNAME

    # patch the dependencies installation method so it skips all subprocessing but actually
    # creates the directory, to simplify testing
    builder._install_dependencies = lambda dirpath: dirpath.mkdir(exist_ok=True)

    # first run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Dependencies directory not found",
        "Installing dependencies",
    )

    # directory created and packages installed
    assert staging_venv_dir.exists()

    # installation directory copied to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]

    # avoid the file to be read successfully
    (tmp_path / const.DEPENDENCIES_HASH_FILENAME).write_bytes(b"\xc3\x28")  # invalid UTF8

    # second run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Problems reading the dependencies hash file: "
        "'utf-8' codec can't decode byte 0xc3 in position 0: invalid continuation byte",
        "Installing dependencies",
    )

    # directory created and packages installed *again*
    assert staging_venv_dir.exists()

    # installation directory copied *again* to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]


@pytest.mark.parametrize(
    ("new_reqs_content", "new_pypackages", "new_pybinaries", "new_charmlibdeps"),
    [
        ("ops==2", None, None, None),
        (None, ["foo2", "bar"], None, None),
        (None, None, ["otherbinthing"], None),
        (None, None, None, {"bazooka"}),
    ],
)
def test_build_dependencies_no_reused_different_dependencies(
    tmp_path,
    assert_output,
    new_reqs_content,
    new_pypackages,
    new_pybinaries,
    new_charmlibdeps,
):
    """Dependencies are built again because changed from previous run."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    # prepare some dependencies for the first call (and some content for the second one)
    reqs_file = tmp_path / "requirements.txt"
    reqs_file.write_text("ops==1")
    requirements = [reqs_file]
    python_packages = ["foo", "bar"]
    binary_python_packages = ["binthing"]
    charmlib_deps = {"baz"}

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
        binary_python_packages=binary_python_packages,
        python_packages=python_packages,
        requirements=requirements,
    )
    builder.charmlib_deps = charmlib_deps
    staging_venv_dir = tmp_path / const.STAGING_VENV_DIRNAME

    # patch the dependencies installation method so it skips all subprocessing but actually
    # creates the directory, to simplify testing
    builder._install_dependencies = lambda dirpath: dirpath.mkdir(exist_ok=True)

    # first run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Dependencies directory not found",
        "Installing dependencies",
    )

    # directory created and packages installed
    assert staging_venv_dir.exists()

    # installation directory copied to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]

    # for the second call, default new dependencies to first ones so only one is changed at a time
    if new_reqs_content is not None:
        reqs_file.write_text(new_reqs_content)
    if new_pypackages is None:
        new_pypackages = python_packages
    if new_pybinaries is None:
        new_pybinaries = binary_python_packages
    if new_charmlibdeps is None:
        new_charmlibdeps = charmlib_deps

    # second run with other dependencies!
    builder.binary_python_packages = new_pybinaries
    builder.python_packages = new_pypackages
    builder.charmlib_deps = new_charmlibdeps
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output("Handling dependencies", "Installing dependencies")

    # directory created and packages installed *again*
    assert staging_venv_dir.exists()

    # installation directory copied *again* to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]


def test_build_dependencies_reused(tmp_path, assert_output):
    """Happy case to reuse dependencies from last run."""
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    reqs_file = tmp_path / "reqs.txt"
    reqs_file.touch()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
        binary_python_packages=[],
        python_packages=[],
        requirements=[reqs_file],
    )
    staging_venv_dir = tmp_path / const.STAGING_VENV_DIRNAME

    # patch the dependencies installation method so it skips all subprocessing but actually
    # creates the directory, to simplify testing; note that we specifically are calling mkdir
    # to fail if the directory is already there, so we ensure it is called once
    builder._install_dependencies = lambda dirpath: dirpath.mkdir(exist_ok=False)

    # first run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Dependencies directory not found",
        "Installing dependencies",
    )

    # directory created and packages installed
    assert staging_venv_dir.exists()

    # installation directory copied to the build directory
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]

    # second run!
    with patch("shutil.copytree") as mock_copytree:
        builder.handle_dependencies()
    assert_output(
        "Handling dependencies",
        "Reusing installed dependencies, they are equal to last run ones",
    )

    # installation directory copied *again* to the build directory (this is always done as
    # buildpath is cleaned)
    site_packages_dir = charm_builder._find_venv_site_packages(
        pathlib.Path(const.STAGING_VENV_DIRNAME)
    )
    assert mock_copytree.mock_calls == [call(site_packages_dir, build_dir / const.VENV_DIRNAME)]


# -- tests about juju ignore


def test_builder_without_jujuignore(tmp_path):
    """Without a .jujuignore we still have a default set of ignores"""
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
    )
    ignore = builder._load_juju_ignore()
    assert ignore.match("/.git", is_dir=True)
    assert ignore.match("/build", is_dir=True)
    assert not ignore.match("myfile.py", is_dir=False)


def test_builder_with_jujuignore(tmp_path):
    """With a .jujuignore we will include additional ignores."""
    metadata = tmp_path / const.METADATA_FILENAME
    metadata.write_text("name: crazycharm")
    build_dir = tmp_path / const.BUILD_DIRNAME
    build_dir.mkdir()
    with (tmp_path / ".jujuignore").open("w", encoding="utf-8") as ignores:
        ignores.write("*.py\n/h\xef.txt\n")

    builder = CharmBuilder(
        builddir=tmp_path,
        installdir=build_dir,
        entrypoint=pathlib.Path("whatever"),
    )
    ignore = builder._load_juju_ignore()
    assert ignore.match("/.git", is_dir=True)
    assert ignore.match("/build", is_dir=True)
    assert ignore.match("myfile.py", is_dir=False)
    assert not ignore.match("hi.txt", is_dir=False)
    assert ignore.match("h\xef.txt", is_dir=False)
    assert not ignore.match("myfile.c", is_dir=False)


def test_builder_arguments_defaults(tmp_path):
    """The arguments passed to the cli must be correctly parsed."""

    def mock_build_charm(self):
        assert self.builddir == pathlib.Path("builddir")
        assert self.installdir == pathlib.Path("installdir")
        assert self.entrypoint == pathlib.Path("src/charm.py")
        assert self.requirement_paths == []
        sys.exit(42)

    fake_argv = ["cmd", "--builddir", "builddir", "--installdir", "installdir"]
    with patch.object(sys, "argv", fake_argv):
        with patch("charmcraft.charm_builder.CharmBuilder.build_charm", new=mock_build_charm):
            with patch("charmcraft.charm_builder.collect_charmlib_pydeps") as mock_collect_pydeps:
                with pytest.raises(SystemExit) as raised:
                    charm_builder.main()
        assert raised.value.code == 42
    mock_collect_pydeps.assert_called_with(pathlib.Path("builddir"))


def test_builder_arguments_full(tmp_path):
    """The arguments passed to the cli must be correctly parsed."""

    def mock_build_charm(self):
        assert self.builddir == pathlib.Path("builddir")
        assert self.installdir == pathlib.Path("installdir")
        assert self.entrypoint == pathlib.Path("src/charm.py")
        assert self.requirement_paths == [
            pathlib.Path("reqs1.txt"),
            pathlib.Path("reqs2.txt"),
        ]
        sys.exit(42)

    fake_argv = ["cmd", "--builddir", "builddir", "--installdir", "installdir"]
    fake_argv += ["-rreqs1.txt", "--requirement", "reqs2.txt"]
    with patch.object(sys, "argv", fake_argv):
        with patch("charmcraft.charm_builder.CharmBuilder.build_charm", new=mock_build_charm):
            with patch("charmcraft.charm_builder.collect_charmlib_pydeps") as mock_collect_pydeps:
                with pytest.raises(SystemExit) as raised:
                    charm_builder.main()
        assert raised.value.code == 42
    mock_collect_pydeps.assert_called_with(pathlib.Path("builddir"))


# --- subprocess runner tests


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_processrun_base(assert_output):
    """Basic execution."""
    cmd = ["echo", "HELO"]
    _process_run(cmd)
    assert_output(
        "Running external command ['echo', 'HELO']",
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_processrun_stdout_logged(assert_output):
    """The standard output is logged in debug."""
    cmd = ["echo", "HELO"]
    _process_run(cmd)
    assert_output(
        "Running external command ['echo', 'HELO']",
        "   :: HELO",
    )


def test_processrun_stderr_logged(assert_output):
    """The standard error is logged in debug."""
    cmd = [sys.executable, "-c", "import sys; print('weird, huh?', file=sys.stderr)"]
    _process_run(cmd)
    assert_output(
        "Running external command " + str(cmd),
        "   :: weird, huh?",
    )


def test_processrun_failed():
    """It's logged in error if subprocess is run but ends with return code not zero."""
    cmd = [sys.executable, "-c", "exit(3)"]
    with pytest.raises(RuntimeError) as cm:
        _process_run(cmd)
    assert str(cm.value) == f"Subprocess command {cmd} execution failed with retcode 3"


def test_processrun_crashed(tmp_path):
    """It's logged in error if the subprocess fails to even run."""
    nonexistent = tmp_path / "whatever"
    cmd = [str(nonexistent)]
    with pytest.raises(RuntimeError) as cm:
        _process_run(cmd)

    # get a real exception to build the message as its internal text varies across OSes
    try:
        subprocess.run([nonexistent], check=True)
    except Exception as exc:
        exc_text = repr(exc)

    assert str(cm.value) == f"Subprocess command {cmd} execution crashed: {exc_text}"


# --- helper tests


@pytest.mark.parametrize(
    ("platform", "result"),
    [
        ("win32", "/basedir/Scripts/cmd.exe"),
        ("linux", "/basedir/bin/cmd"),
        ("darwin", "/basedir/bin/cmd"),
    ],
)
def test_find_venv_bin(monkeypatch, platform, result):
    monkeypatch.setattr(sys, "platform", platform)
    basedir = pathlib.Path("/basedir")
    venv_bin = charm_builder._find_venv_bin(basedir, "cmd")
    assert venv_bin.as_posix() == result


@pytest.mark.parametrize(
    ("platform", "result"),
    [
        ("win32", "/basedir/PythonXY/site-packages"),
        ("linux", "/basedir/lib/pythonX.Y/site-packages"),
        ("darwin", "/basedir/lib/pythonX.Y/site-packages"),
    ],
)
def test_find_venv_site_packages(monkeypatch, platform, result):
    monkeypatch.setattr(sys, "platform", platform)
    basedir = pathlib.Path("/basedir")
    with patch("subprocess.check_output", return_value="X Y") as mock_run:
        site_packages_dir = charm_builder._find_venv_site_packages(basedir)
    assert mock_run.mock_calls == [
        call(
            [
                "python3",
                "-c",
                "import sys; v=sys.version_info; print(f'{v.major} {v.minor}')",
            ],
            text=True,
        )
    ]
    assert site_packages_dir.as_posix() == result



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Tests for the OCI Registry related functionality (code in store/registry.py)."""

import base64
import gzip
import hashlib
import io
import json
import pathlib
import sys
import tarfile
from unittest.mock import call, patch

import pytest
import requests
from craft_cli import CraftError

from charmcraft import const
from charmcraft.store import registry
from charmcraft.store.registry import (
    CONFIG_MIMETYPE,
    LAYER_MIMETYPE,
    MANIFEST_V2_MIMETYPE,
    OCTET_STREAM_MIMETYPE,
    ImageHandler,
    LocalDockerdInterface,
    OCIRegistry,
    assert_response_ok,
)

# -- tests for response verifications


def create_response(
    status_code=200, headers=None, raw_content=b"", json_content=None, content_type=None
):
    """Create a fake requests' response."""
    if headers is None:
        headers = {}

    if json_content is not None:
        headers.setdefault("Content-Type", content_type or "application/json")
        content_bytes = json.dumps(json_content).encode("utf8")
    else:
        content_bytes = raw_content

    resp = requests.Response()
    resp.status_code = status_code
    resp.raw = io.BytesIO(content_bytes)
    resp.headers = headers  # not case insensitive, but good enough
    return resp


def test_assert_response_ok_simple_json():
    """Simple case for a good response with JSON content."""
    test_content = {"foo": 2, "bar": 1}
    response = create_response(json_content=test_content)
    result = assert_response_ok(response)
    assert result == test_content


def test_assert_response_ok_not_json():
    """A good non-json response."""
    response = create_response(raw_content=b"stuff")
    result = assert_response_ok(response)
    assert result is None


def test_assert_response_ok_different_status():
    """A good response with a different status code."""
    test_content = {"foo": 2, "bar": 1}
    response = create_response(json_content=test_content, status_code=201)
    result = assert_response_ok(response, expected_status=201)
    assert result == test_content


def test_assert_response_errors_in_result():
    """Response is as expected but server flags errors."""
    errors = [{"foo": "bar"}]
    test_content = {"errors": errors}
    response = create_response(json_content=test_content)
    with pytest.raises(CraftError) as cm:
        assert_response_ok(response)
    assert str(cm.value) == f"Response with errors from server: {errors}"


def test_assert_response_bad_status_code_with_json_errors():
    """Different status code than expected, with the server including errors."""
    errors = [{"foo": "bar"}]
    test_content = {"errors": errors}
    response = create_response(status_code=404, json_content=test_content)
    with pytest.raises(CraftError) as cm:
        assert_response_ok(response)
    error = cm.value
    assert str(error) == "Wrong status code from server (expected=200, got=404)"
    assert error.details == f"errors={errors} headers={{'Content-Type': 'application/json'}}"


def test_assert_response_bad_status_code_with_extra_json_errors():
    """The server still including errors, weird content type."""
    errors = [{"foo": "bar"}]
    test_content = {"errors": errors}
    response = create_response(
        status_code=404,
        json_content=test_content,
        content_type="application/json;stuff",
    )
    with pytest.raises(CraftError) as cm:
        assert_response_ok(response)
    error = cm.value
    assert str(error) == "Wrong status code from server (expected=200, got=404)"
    assert error.details == f"errors={errors} headers={{'Content-Type': 'application/json;stuff'}}"


def test_assert_response_bad_status_code_blind():
    """Different status code than expected, no more info."""
    response = create_response(status_code=500, raw_content=b"stuff")
    with pytest.raises(CraftError) as cm:
        assert_response_ok(response)
    error = cm.value
    assert str(error) == "Wrong status code from server (expected=200, got=500)"
    assert error.details == "errors=None headers={}"


# -- tests for OCIRegistry auth & hit helpers


def test_auth_simple(responses):
    """Simple authentication."""
    responses.add(
        responses.GET,
        "https://auth.fakereg.com?service=test-service&scope=test-scope",
        json={"token": "test-token"},
    )

    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    auth_info = {
        "realm": "https://auth.fakereg.com",
        "service": "test-service",
        "scope": "test-scope",
    }
    token = ocireg._authenticate(auth_info)
    assert token == "test-token"
    sent_auth_header = responses.calls[0].request.headers.get("Authorization")
    assert sent_auth_header is None


def test_auth_with_credentials(emitter, responses):
    """Authenticate passing credentials."""
    responses.add(
        responses.GET,
        "https://auth.fakereg.com?service=test-service&scope=test-scope",
        json={"token": "test-token"},
    )

    ocireg = OCIRegistry(
        "https://fakereg.com",
        "test-image",
        username="test-user",
        password="test-password",
    )
    auth_info = {
        "realm": "https://auth.fakereg.com",
        "service": "test-service",
        "scope": "test-scope",
    }
    token = ocireg._authenticate(auth_info)
    assert token == "test-token"
    sent_auth_header = responses.calls[0].request.headers.get("Authorization")
    expected_encoded = base64.b64encode(b"test-user:test-password")
    assert sent_auth_header == "Basic " + expected_encoded.decode("ascii")

    # generic auth indication is logged but NOT the credentials
    expected = f"Authenticating! {auth_info}"
    emitter.assert_trace(expected)


def test_auth_with_just_username(responses):
    """Authenticate passing credentials."""
    responses.add(
        responses.GET,
        "https://auth.fakereg.com?service=test-service&scope=test-scope",
        json={"token": "test-token"},
    )

    ocireg = OCIRegistry("https://fakereg.com", "test-image", username="test-user")
    auth_info = {
        "realm": "https://auth.fakereg.com",
        "service": "test-service",
        "scope": "test-scope",
    }
    token = ocireg._authenticate(auth_info)
    assert token == "test-token"
    sent_auth_header = responses.calls[0].request.headers.get("Authorization")
    expected_encoded = base64.b64encode(b"test-user:")
    assert sent_auth_header == "Basic " + expected_encoded.decode("ascii")


def test_hit_simple_initial_auth_ok(emitter, responses):
    """Simple GET with auth working at once."""
    # set the Registry with an initial token
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    ocireg.auth_token = "some auth token"

    # fake a 200 response
    responses.add(responses.GET, "https://fakereg.com/api/stuff")

    # try it
    response = ocireg._hit("GET", "https://fakereg.com/api/stuff")
    assert response == responses.calls[0].response

    # verify it authed ok
    sent_auth_header = responses.calls[0].request.headers.get("Authorization")
    assert sent_auth_header == "Bearer some auth token"

    # logged what it did
    expected = "Hitting the registry: GET https://fakereg.com/api/stuff"
    emitter.assert_trace(expected)


def test_hit_simple_re_auth_ok(responses):
    """Simple GET but needing to re-authenticate."""
    # set the Registry
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    ocireg.auth_token = "some auth token"

    # need to set up two responses!
    # - the 401 response with the proper info to re-auth
    # - the request that actually works
    headers = {
        "Www-Authenticate": (
            'Bearer realm="https://auth.fakereg.com/token",'
            'service="https://fakereg.com",scope="repository:library/stuff:pull"'
        )
    }
    responses.add(responses.GET, "https://fakereg.com/api/stuff", headers=headers, status=401)
    responses.add(responses.GET, "https://fakereg.com/api/stuff")

    # try it, isolating the re-authentication (tested separately above)
    with patch.object(ocireg, "_authenticate") as mock_auth:
        mock_auth.return_value = "new auth token"
        response = ocireg._hit("GET", "https://fakereg.com/api/stuff")
    assert response == responses.calls[1].response
    mock_auth.assert_called_with(
        {
            "realm": "https://auth.fakereg.com/token",
            "scope": "repository:library/stuff:pull",
            "service": "https://fakereg.com",
        }
    )

    # verify it authed ok both times, with corresponding tokens, and that it stored the new one
    sent_auth_header = responses.calls[0].request.headers.get("Authorization")
    assert sent_auth_header == "Bearer some auth token"
    sent_auth_header = responses.calls[1].request.headers.get("Authorization")
    assert sent_auth_header == "Bearer new auth token"
    assert ocireg.auth_token == "new auth token"


def test_hit_simple_re_auth_problems(responses):
    """Bad response from the re-authentication process."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")

    # set only one response, a 401 which is broken and all will end there
    headers = {"Www-Authenticate": "broken header"}
    responses.add(responses.GET, "https://fakereg.com/api/stuff", headers=headers, status=401)

    # try it, isolating the re-authentication (tested separately above)
    expected = (
        "Bad 401 response: Bearer not found; headers: {.*'Www-Authenticate': 'broken header'.*}"
    )
    with pytest.raises(CraftError, match=expected):
        ocireg._hit("GET", "https://fakereg.com/api/stuff")


def test_hit_different_method(responses):
    """Simple request using something else than GET."""
    # set the Registry with an initial token
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    ocireg.auth_token = "some auth token"

    # fake a 200 response
    responses.add(responses.POST, "https://fakereg.com/api/stuff")

    # try it
    response = ocireg._hit("POST", "https://fakereg.com/api/stuff")
    assert response == responses.calls[0].response


def test_hit_including_headers(responses):
    """A request including more headers."""
    # set the Registry with an initial token
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    ocireg.auth_token = "some auth token"

    # fake a 200 response
    responses.add(responses.POST, "https://fakereg.com/api/stuff")

    # try it
    response = ocireg._hit("POST", "https://fakereg.com/api/stuff", headers={"FOO": "bar"})
    assert response == responses.calls[0].response

    # check that it sent the requested header AND the automatic auth one
    sent_headers = responses.calls[0].request.headers
    assert sent_headers.get("FOO") == "bar"
    assert sent_headers.get("Authorization") == "Bearer some auth token"


def test_hit_extra_parameters(responses):
    """The request can include extra parameters."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")

    # fake a 200 response
    responses.add(responses.PUT, "https://fakereg.com/api/stuff")

    # try it
    response = ocireg._hit("PUT", "https://fakereg.com/api/stuff", data=b"test-payload")
    assert response == responses.calls[0].response
    assert responses.calls[0].request.body == b"test-payload"


def test_hit_no_log(emitter, responses):
    """Simple request but avoiding log."""
    # set the Registry with an initial token
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    ocireg.auth_token = "some auth token"

    # fake a 200 response
    responses.add(responses.PUT, "https://fakereg.com/api/stuff")

    # try it
    ocireg._hit("PUT", "https://fakereg.com/api/stuff", log=False)

    # nothing shown!
    emitter.assert_interactions(None)


# -- tests for other OCIRegistry helpers: checkers if stuff uploaded


def test_ociregistry_is_manifest_uploaded():
    """Check the simple call with correct path to the generic verifier."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    with patch.object(ocireg, "_is_item_already_uploaded") as mock_verifier:
        mock_verifier.return_value = "whatever"
        result = ocireg.is_manifest_already_uploaded("test-reference")
    assert result == "whatever"
    url = "https://fakereg.com/v2/test-image/manifests/test-reference"
    mock_verifier.assert_called_with(url)


def test_ociregistry_is_blob_uploaded():
    """Check the simple call with correct path to the generic verifier."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    with patch.object(ocireg, "_is_item_already_uploaded") as mock_verifier:
        mock_verifier.return_value = "whatever"
        result = ocireg.is_blob_already_uploaded("test-reference")
    assert result == "whatever"
    url = "https://fakereg.com/v2/test-image/blobs/test-reference"
    mock_verifier.assert_called_with(url)


def test_ociregistry_is_item_uploaded_simple_yes(responses):
    """Simple case for the item already uploaded."""
    ocireg = OCIRegistry("http://fakereg.com/", "test-image")
    url = "http://fakereg.com/v2/test-image/stuff/some-reference"
    responses.add(responses.HEAD, url)

    # try it
    result = ocireg._is_item_already_uploaded(url)
    assert result is True


def test_ociregistry_is_item_uploaded_simple_no(responses):
    """Simple case for the item NOT already uploaded."""
    ocireg = OCIRegistry("http://fakereg.com/", "test-image")
    url = "http://fakereg.com/v2/test-image/stuff/some-reference"
    responses.add(responses.HEAD, url, status=404)

    # try it
    result = ocireg._is_item_already_uploaded(url)
    assert result is False


@pytest.mark.parametrize("redir_status", [302, 307])
def test_ociregistry_is_item_uploaded_redirect(responses, redir_status):
    """The verification is redirected to somewhere else."""
    ocireg = OCIRegistry("http://fakereg.com/", "test-image")
    url1 = "http://fakereg.com/v2/test-image/stuff/some-reference"
    url2 = "http://fakereg.com/real-check/test-image/stuff/some-reference"
    responses.add(responses.HEAD, url1, status=redir_status, headers={"Location": url2})
    responses.add(responses.HEAD, url2, status=200)

    # try it
    result = ocireg._is_item_already_uploaded(url1)
    assert result is True


def test_ociregistry_is_item_uploaded_strange_response(responses, emitter):
    """Unexpected response."""
    ocireg = OCIRegistry("http://fakereg.com/", "test-image")
    url = "http://fakereg.com/v2/test-image/stuff/some-reference"
    responses.add(responses.HEAD, url, status=400, headers={"foo": "bar"})

    # try it
    result = ocireg._is_item_already_uploaded(url)
    assert result is False
    expected = (
        "Bad response when checking for uploaded "
        "'http://fakereg.com/v2/test-image/stuff/some-reference': 400 "
        "(headers={'Content-Type': 'text/plain', 'foo': 'bar'})"
    )
    emitter.assert_debug(expected)


# -- test for the OCIRegistry manifest upload


def test_ociregistry_upload_manifest_v2(responses, emitter):
    """Upload a V2 manifest."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")

    url = "https://fakereg.com/v2/test-image/manifests/test-reference"
    responses.add(responses.PUT, url, status=201)

    # try it
    raw_manifest_data = "test-manifest"
    ocireg.upload_manifest(raw_manifest_data, "test-reference")

    # check logs
    emitter.assert_progress("Uploading manifest with reference test-reference")
    emitter.assert_progress("Manifest uploaded OK")

    # check header and data sent
    assert responses.calls[0].request.headers["Content-Type"] == MANIFEST_V2_MIMETYPE
    assert responses.calls[0].request.body == raw_manifest_data.encode("ascii")


# -- tests for the OCIRegistry blob upload


def test_ociregistry_upload_blob_complete(tmp_path, emitter, responses, monkeypatch):
    """Complete upload of a binary to the registry."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    base_url = "https://fakereg.com/v2/test-image/"

    # fake the first initial response
    pump_url_1 = base_url + "fakeurl-1"
    responses.add(
        responses.POST,
        base_url + "blobs/uploads/",
        status=202,
        headers={"Location": pump_url_1, "Range": "0-0"},
    )

    # and the intermediate ones, chained
    pump_url_2 = base_url + "fakeurl-2"
    pump_url_3 = base_url + "fakeurl-3"
    pump_url_4 = base_url + "fakeurl-4"
    responses.add(responses.PATCH, pump_url_1, status=202, headers={"Location": pump_url_2})
    responses.add(responses.PATCH, pump_url_2, status=202, headers={"Location": pump_url_3})
    responses.add(responses.PATCH, pump_url_3, status=202, headers={"Location": pump_url_4})

    # finally, the closing url
    responses.add(
        responses.PUT,
        base_url + "fakeurl-4&digest=test-digest",
        status=201,
        headers={"Docker-Content-Digest": "test-digest"},
    )

    # prepare a fake content that will be pushed in 3 parts
    monkeypatch.setattr(registry, "CHUNK_SIZE", 3)
    bytes_source = tmp_path / "testfile"
    bytes_source.write_text("abcdefgh")

    # call!
    ocireg.upload_blob(bytes_source, 8, "test-digest")

    # check all the sent headers
    expected_headers_per_request = [
        {},  # nothing special in the initial one
        {
            "Content-Length": "3",
            "Content-Range": "0-3",
            "Content-Type": OCTET_STREAM_MIMETYPE,
        },
        {
            "Content-Length": "3",
            "Content-Range": "3-6",
            "Content-Type": OCTET_STREAM_MIMETYPE,
        },
        {
            "Content-Length": "2",
            "Content-Range": "6-8",
            "Content-Type": OCTET_STREAM_MIMETYPE,
        },
        {"Content-Length": "0", "Connection": "close"},  # closing
    ]
    for idx, expected_headers in enumerate(expected_headers_per_request):
        sent_headers = responses.calls[idx].request.headers
        for key, value in expected_headers.items():
            assert sent_headers.get(key) == value

    emitter.assert_interactions(
        [
            call("progress", "Getting URL to push the blob"),
            call(
                "trace",
                "Hitting the registry: POST https://fakereg.com/v2/test-image/blobs/uploads/",
            ),
            call("progress", "Got upload URL ok with range 0-0"),
            call("progress_bar", "Uploading...", 8),
            call("advance", 3),
            call("advance", 3),
            call("advance", 2),
            call("progress", "Closing the upload"),
            call(
                "trace",
                (
                    "Hitting the registry: PUT "
                    "https://fakereg.com/v2/test-image/fakeurl-4&digest=test-digest"
                ),
            ),
            call("progress", "Upload finished OK"),
        ]
    )


def test_ociregistry_upload_blob_bad_initial_response(responses):
    """Bad initial response when starting to upload."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    base_url = "https://fakereg.com/v2/test-image/"

    # fake the first initial response with problems
    responses.add(responses.POST, base_url + "blobs/uploads/", status=500)

    # call!
    msg = r"Wrong status code from server \(expected=202, got=500\).*"
    with pytest.raises(CraftError, match=msg):
        ocireg.upload_blob("test-filepath", 8, "test-digest")


def test_ociregistry_upload_blob_bad_upload_range(responses):
    """Received a broken range info."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    base_url = "https://fakereg.com/v2/test-image/"

    # fake the first initial response with problems
    responses.add(
        responses.POST,
        base_url + "blobs/uploads/",
        status=202,
        headers={"Location": "test-next-url", "Range": "9-9"},
    )

    # call!
    with pytest.raises(CraftError) as cm:
        ocireg.upload_blob("test-filepath", 8, "test-digest")
    error = cm.value
    assert str(error) == "Server error: bad range received"
    assert error.details == "Range='9-9'"


def test_ociregistry_upload_blob_resumed(tmp_path, emitter, responses):
    """The upload is resumed after server indication to do so."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    base_url = "https://fakereg.com/v2/test-image/"

    # fake the first initial response, indicating that the store has already the first 5 bytes
    pump_url_1 = base_url + "fakeurl-1"
    responses.add(
        responses.POST,
        base_url + "blobs/uploads/",
        status=202,
        headers={"Location": pump_url_1, "Range": "0-4"},
    )  # has bytes in position 0, 1, 2, 3 & 4

    # and the intermediate one
    pump_url_2 = base_url + "fakeurl-2"
    responses.add(responses.PATCH, pump_url_1, status=202, headers={"Location": pump_url_2})

    # finally, the closing url
    responses.add(
        responses.PUT,
        base_url + "fakeurl-2&digest=test-digest",
        status=201,
        headers={"Docker-Content-Digest": "test-digest"},
    )

    # prepare a fake content
    bytes_source = tmp_path / "testfile"
    bytes_source.write_text("abcdefgh")

    # call!
    ocireg.upload_blob(bytes_source, 8, "test-digest")

    # check all the sent headers
    expected_headers_per_request = [
        {},  # nothing special in the initial one
        {
            "Content-Length": "3",
            "Content-Range": "5-8",
            "Content-Type": OCTET_STREAM_MIMETYPE,
        },
        {"Content-Length": "0", "Connection": "close"},  # closing
    ]
    for idx, expected_headers in enumerate(expected_headers_per_request):
        sent_headers = responses.calls[idx].request.headers
        for key, value in expected_headers.items():
            assert sent_headers.get(key) == value

    emitter.assert_interactions(
        [
            call("progress", "Getting URL to push the blob"),
            call(
                "trace",
                "Hitting the registry: POST https://fakereg.com/v2/test-image/blobs/uploads/",
            ),
            call("progress", "Got upload URL ok with range 0-4"),
            call("progress_bar", "Uploading...", 8),
            call("advance", 5),
            call("advance", 3),
            call("progress", "Closing the upload"),
            call(
                "trace",
                (
                    "Hitting the registry: PUT "
                    "https://fakereg.com/v2/test-image/fakeurl-2&digest=test-digest"
                ),
            ),
            call("progress", "Upload finished OK"),
        ]
    )


def test_ociregistry_upload_blob_bad_response_middle(tmp_path, responses, monkeypatch):
    """Bad response from the server when pumping bytes."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    base_url = "https://fakereg.com/v2/test-image/"

    # fake the first initial response
    pump_url_1 = base_url + "fakeurl-1"
    responses.add(
        responses.POST,
        base_url + "blobs/uploads/",
        status=202,
        headers={"Location": pump_url_1, "Range": "0-0"},
    )

    # and the intermediate ones, chained, with a crash
    pump_url_2 = base_url + "fakeurl-2"
    responses.add(responses.PATCH, pump_url_1, status=202, headers={"Location": pump_url_2})
    responses.add(responses.PATCH, pump_url_2, status=504)

    # prepare a fake content that will be pushed in 3 parts
    monkeypatch.setattr(registry, "CHUNK_SIZE", 3)
    bytes_source = tmp_path / "testfile"
    bytes_source.write_text("abcdefgh")

    # call!
    msg = r"Wrong status code from server \(expected=202, got=504\).*"
    with pytest.raises(CraftError, match=msg):
        ocireg.upload_blob(bytes_source, 8, "test-digest")


def test_ociregistry_upload_blob_bad_response_closing(tmp_path, responses):
    """Bad response from the server when closing the upload."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    base_url = "https://fakereg.com/v2/test-image/"

    # fake the first initial response
    pump_url_1 = base_url + "fakeurl-1"
    responses.add(
        responses.POST,
        base_url + "blobs/uploads/",
        status=202,
        headers={"Location": pump_url_1, "Range": "0-0"},
    )

    # and the intermediate one
    pump_url_2 = base_url + "fakeurl-2"
    responses.add(responses.PATCH, pump_url_1, status=202, headers={"Location": pump_url_2})

    # finally, the closing url, crashing
    responses.add(responses.PUT, base_url + "fakeurl-2&digest=test-digest", status=502)

    # prepare a fake content
    bytes_source = tmp_path / "testfile"
    bytes_source.write_text("abcdefgh")

    # call!
    msg = r"Wrong status code from server \(expected=201, got=502\).*"
    with pytest.raises(CraftError, match=msg):
        ocireg.upload_blob(bytes_source, 8, "test-digest")


def test_ociregistry_upload_blob_bad_final_digest(tmp_path, responses):
    """Bad digest from server after closing the upload."""
    ocireg = OCIRegistry("https://fakereg.com", "test-image")
    base_url = "https://fakereg.com/v2/test-image/"

    # fake the first initial response
    pump_url_1 = base_url + "fakeurl-1"
    responses.add(
        responses.POST,
        base_url + "blobs/uploads/",
        status=202,
        headers={"Location": pump_url_1, "Range": "0-0"},
    )

    # and the intermediate one
    pump_url_2 = base_url + "fakeurl-2"
    responses.add(responses.PATCH, pump_url_1, status=202, headers={"Location": pump_url_2})

    # finally, the closing url, bad digest
    responses.add(
        responses.PUT,
        base_url + "fakeurl-2&digest=test-digest",
        status=201,
        headers={"Docker-Content-Digest": "somethingelse"},
    )

    # prepare a fake content
    bytes_source = tmp_path / "testfile"
    bytes_source.write_text("abcdefgh")

    # call!
    msg = "Server error: the upload is corrupted"
    with pytest.raises(CraftError, match=msg):
        ocireg.upload_blob(bytes_source, 8, "test-digest")


# -- tests for the ImageHandler helpers and functionalities


def test_localdockerinterface_get_info_by_id_ok(responses, emitter):
    """Get image info ok."""
    test_image_info = {"some": "stuff"}
    responses.add(
        responses.GET,
        LocalDockerdInterface.dockerd_socket_baseurl + "/images/test-id/json",
        json=test_image_info,
    )
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_id("test-id")
    assert resp == test_image_info

    emitter.assert_interactions(None)


def test_localdockerinterface_get_info_by_id_not_found(responses, emitter):
    """Get image info for something that is not there."""
    # return 404, which means that the image was not found
    responses.add(
        responses.GET,
        LocalDockerdInterface.dockerd_socket_baseurl + "/images/test-id/json",
        status=404,
    )
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_id("test-id")
    assert resp is None

    emitter.assert_interactions(None)


def test_localdockerinterface_get_info_by_id_bad_response(responses, emitter):
    """Docker answered badly when checking for the image."""
    # weird dockerd behaviour
    responses.add(
        responses.GET,
        LocalDockerdInterface.dockerd_socket_baseurl + "/images/test-id/json",
        status=500,
    )
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_id("test-id")
    assert resp is None

    emitter.assert_debug("Bad response when validating local image: 500")


def test_localdockerinterface_get_info_by_id_disconnected(emitter, responses):
    """No daemon to talk to (see responses used as fixture but no listening)."""
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_id("test-id")
    assert resp is None

    emitter.assert_debug(
        "Cannot connect to /var/run/docker.sock , please ensure dockerd is running."
    )


def test_localdockerinterface_get_info_by_digest_ok(responses, emitter):
    """Get image info ok."""
    test_image_info_1 = {"some": "stuff", "RepoDigests": ["name @ sha256:test-digest", "other"]}
    test_image_info_2 = {"some": "stuff", "RepoDigests": ["foo", "bar"]}
    test_search_respoonse = [test_image_info_1, test_image_info_2]
    responses.add(
        responses.GET,
        LocalDockerdInterface.dockerd_socket_baseurl + "/images/json",
        json=test_search_respoonse,
    )
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_digest("sha256:test-digest")
    assert resp == test_image_info_1

    emitter.assert_interactions(None)


def test_localdockerinterface_get_info_by_digest_not_found(responses, emitter):
    """Get image info for something that is not there."""
    test_image_info_1 = {"some": "stuff", "RepoDigests": ["other"]}
    test_image_info_2 = {"some": "stuff", "RepoDigests": ["foo", "bar"]}
    test_search_respoonse = [test_image_info_1, test_image_info_2]
    responses.add(
        responses.GET,
        LocalDockerdInterface.dockerd_socket_baseurl + "/images/json",
        json=test_search_respoonse,
    )
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_digest("sha256:test-digest")
    assert resp is None

    emitter.assert_interactions(None)


def test_localdockerinterface_get_info_by_digest_none_digest(responses, emitter):
    """Get image info for something that is not there."""
    test_image_info_1 = {"some": "stuff", "RepoDigests": None}
    test_search_respoonse = [test_image_info_1]
    responses.add(
        responses.GET,
        LocalDockerdInterface.dockerd_socket_baseurl + "/images/json",
        json=test_search_respoonse,
    )
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_digest("sha256:test-digest")
    assert resp is None

    emitter.assert_interactions(None)


def test_localdockerinterface_get_info_by_digest_bad_response(responses, emitter):
    """Docker answered badly when checking for the image."""
    # weird dockerd behaviour
    responses.add(
        responses.GET,
        LocalDockerdInterface.dockerd_socket_baseurl + "/images/json",
        status=500,
    )
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_digest("sha256:test-digest")
    assert resp is None

    emitter.assert_debug("Bad response when validating local image: 500")


def test_localdockerinterface_get_info_by_digest_disconnected(emitter, responses):
    """No daemon to talk to (see responses used as fixture but no listening)."""
    ldi = LocalDockerdInterface()
    resp = ldi.get_image_info_from_digest("sha256:test-digest")
    assert resp is None

    emitter.assert_debug(
        "Cannot connect to /var/run/docker.sock , please ensure dockerd is running."
    )


def test_localdockerinterface_get_streamed_content(responses):
    """Get the content streamed."""

    class AuditableBufferedReader(io.BufferedReader):
        """BufferedReader that records the size of each reading."""

        _test_read_chunks = []

        def read(self, size):
            self._test_read_chunks.append(size)
            return super().read(size)

    test_content = AuditableBufferedReader(io.BytesIO(b"123456789"))
    responses.add(
        responses.GET,
        LocalDockerdInterface.dockerd_socket_baseurl + "/images/test-id/get",
        body=test_content,
    )
    ldi = LocalDockerdInterface()
    resp = ldi.get_streamed_image_content("test-id")
    assert test_content._test_read_chunks == []

    chunk_size = 5
    streamed = resp.iter_content(chunk_size)
    assert next(streamed) == b"12345"
    assert test_content._test_read_chunks == [chunk_size]
    assert next(streamed) == b"6789"
    assert test_content._test_read_chunks == [chunk_size, chunk_size]
    with pytest.raises(StopIteration):
        next(streamed)


class FakeRegistry:
    """A fake registry to mimic behaviour of the real one and record actions."""

    def __init__(self, image_name=None):
        self.image_name = image_name
        self.stored_manifests = {}
        self.stored_blobs = {}

    def is_manifest_already_uploaded(self, reference):
        return reference in self.stored_manifests

    def upload_manifest(self, content, reference, multiple_manifest=False):
        self.stored_manifests[reference] = (content, multiple_manifest)

    def get_manifest(self, reference):
        return self.stored_manifests[reference]

    def is_blob_already_uploaded(self, reference):
        return reference in self.stored_blobs

    def upload_blob(self, filepath, size, digest):
        self.stored_blobs[digest] = (pathlib.Path(filepath).read_bytes(), size)


class FakeDockerd:
    """A fake dockerd interface to mimic behaviour of the real one."""

    def __init__(self, image_id, image_info, image_content):
        self.image_info = image_info
        self.image_content = image_content
        self.used_id = image_id

    def get_streamed_image_content(self, image_id):
        assert image_id == self.used_id

        class FakeResponse:
            def __init__(self, content):
                self.content = io.BytesIO(content)

            def iter_content(self, chunk_size):
                while True:
                    chunk = self.content.read(chunk_size)
                    if not chunk:
                        break
                    yield chunk

        return FakeResponse(self.image_content)


def test_imagehandler_check_in_registry_yes():
    """Check if an image is in the registry and find it."""
    fake_registry = FakeRegistry()
    fake_registry.stored_manifests["test-reference"] = (
        None,
        "test-digest",
        "test-manifest",
    )

    im = ImageHandler(fake_registry)
    result = im.check_in_registry("test-reference")
    assert result is True


def test_imagehandler_check_in_registry_no():
    """Check if an image is in the registry and don't find it."""
    fake_registry = FakeRegistry()

    im = ImageHandler(fake_registry)
    result = im.check_in_registry("test-reference")
    assert result is False


def test_imagehandler_extract_file_simple(tmp_path, emitter):
    """Extract a file from the tarfile and gets its info."""
    # create a tar file with one file inside
    test_content = b"test content for the sample file"
    sample_file = tmp_path / "testfile.txt"
    sample_file.write_bytes(test_content)
    tar_filepath = tmp_path / "testfile.tar"
    with tarfile.open(tar_filepath, "w") as tar:
        tar.add(sample_file, "testfile.txt")

    im = ImageHandler("registry")
    with tarfile.open(tar_filepath, "r") as tar:
        tmp_filepath, size, digest = im._extract_file(tar, "testfile.txt")

    assert size == len(test_content)
    assert digest == "sha256:" + hashlib.sha256(test_content).hexdigest()
    assert pathlib.Path(tmp_filepath).read_bytes() == test_content

    emitter.assert_progress("Extracting file 'testfile.txt' from local tar (compress=False)")


def test_imagehandler_extract_file_compressed_ok(tmp_path, emitter):
    """Extract a file from the tarfile and gets its info after compressed."""
    # create a tar file with one file inside
    test_content = b"test content for the sample file"
    sample_file = tmp_path / "testfile.txt"
    sample_file.write_bytes(test_content)
    tar_filepath = tmp_path / "testfile.tar"
    with tarfile.open(tar_filepath, "w") as tar:
        tar.add(sample_file, "testfile.txt")

    im = ImageHandler("registry")
    with tarfile.open(tar_filepath, "r") as tar:
        tmp_filepath, size, digest = im._extract_file(tar, "testfile.txt", compress=True)

    compressed_content = pathlib.Path(tmp_filepath).read_bytes()
    assert size == len(compressed_content)
    assert digest == "sha256:" + hashlib.sha256(compressed_content).hexdigest()
    assert gzip.decompress(compressed_content) == test_content

    emitter.assert_progress("Extracting file 'testfile.txt' from local tar (compress=True)")


def test_imagehandler_extract_file_compressed_deterministic(tmp_path, emitter):
    """Different compressions for the same file give the exact same data."""
    # create a tar file with one file inside
    test_content = b"test content for the sample file"
    sample_file = tmp_path / "testfile.txt"
    sample_file.write_bytes(test_content)
    tar_filepath = tmp_path / "testfile.tar"
    with tarfile.open(tar_filepath, "w") as tar:
        tar.add(sample_file, "testfile.txt")

    im = ImageHandler("registry")
    with tarfile.open(tar_filepath, "r") as tar:
        _, _, digest1 = im._extract_file(tar, "testfile.txt", compress=True)
        _, _, digest2 = im._extract_file(tar, "testfile.txt", compress=True)

    assert digest1 == digest2


def test_imagehandler_uploadblob_first_time(emitter, tmp_path):
    """Upload a blob for the first time."""
    tmp_file = tmp_path / "somebinary.dat"
    tmp_file.write_text("testcontent")

    fake_registry = FakeRegistry()

    im = ImageHandler(fake_registry)
    im._upload_blob(str(tmp_file), 20, "superdigest")

    # check it was uploaded
    assert fake_registry.stored_blobs["superdigest"] == (b"testcontent", 20)

    # verify the file is cleaned
    assert not tmp_file.exists()

    emitter.assert_interactions(None)


def test_imagehandler_uploadblob_duplicated(emitter, tmp_path):
    """Upload a blob that was already there."""
    tmp_file = tmp_path / "somebinary.dat"
    tmp_file.write_text("testcontent")

    fake_registry = FakeRegistry()
    # add the entry for the blob, the value is not important
    fake_registry.stored_blobs["superdigest"] = None

    im = ImageHandler(fake_registry)
    im._upload_blob(str(tmp_file), 20, "superdigest")

    # check it was NOT uploaded again
    assert fake_registry.stored_blobs["superdigest"] is None

    # verify the file is cleaned
    assert not tmp_file.exists()

    emitter.assert_progress("Blob was already uploaded")


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_imagehandler_uploadfromlocal_complete(emitter, tmp_path, responses, monkeypatch):
    """Complete process of uploading a local image."""
    # fake an image in disk (a tar file with config, layers, and a manifest)."""
    test_tar_image = tmp_path / "test-image.tar"
    test_tar_config_content = b"fake config for the image"
    test_tar_layer1_content = b"fake first layer content for the image"
    test_tar_layer2_content = b"fake second layer content for the image"
    test_manifest_content = json.dumps(
        [
            {
                "Config": const.JUJU_CONFIG_FILENAME,
                "Layers": ["layer1.bin", "layer2.bin"],
            }
        ]
    ).encode("ascii")
    tar_file = tarfile.TarFile(test_tar_image, "w")
    tar_content = [
        ("manifest.json", test_manifest_content),
        (const.JUJU_CONFIG_FILENAME, test_tar_config_content),
        ("layer1.bin", test_tar_layer1_content),
        ("layer2.bin", test_tar_layer2_content),
    ]
    for name, content in tar_content:
        ti = tarfile.TarInfo(name)
        ti.size = len(content)
        tar_file.addfile(ti, fileobj=io.BytesIO(content))
    tar_file.close()

    # prepare the image info
    image_size = test_tar_image.stat().st_size
    image_id = "test-image-id"
    image_info = {"Size": image_size, "Id": image_id, "foobar": "etc"}
    fakedockerd = FakeDockerd(image_id, image_info, test_tar_image.read_bytes())
    monkeypatch.setattr(registry, "LocalDockerdInterface", lambda: fakedockerd)

    # ensure two reads from that image, so we can properly test progress
    image_read_from_dockerd_size_1 = int(image_size * 0.7)
    image_read_from_dockerd_size_2 = image_size - image_read_from_dockerd_size_1
    monkeypatch.setattr(registry, "CHUNK_SIZE", image_read_from_dockerd_size_1)

    fake_registry = FakeRegistry()
    im = ImageHandler(fake_registry)
    main_call_result = im.upload_from_local(image_info)

    # check the uploaded blobs: first the config (as is), then the layers (compressed)
    (
        uploaded_config,
        uploaded_layer1,
        uploaded_layer2,
    ) = fake_registry.stored_blobs.items()

    (u_config_digest, (u_config_content, u_config_size)) = uploaded_config
    assert u_config_content == test_tar_config_content
    assert u_config_size == len(u_config_content)
    assert u_config_digest == "sha256:" + hashlib.sha256(u_config_content).hexdigest()

    (u_layer1_digest, (u_layer1_content, u_layer1_size)) = uploaded_layer1
    assert gzip.decompress(u_layer1_content) == test_tar_layer1_content
    assert u_layer1_size == len(u_layer1_content)
    assert u_layer1_digest == "sha256:" + hashlib.sha256(u_layer1_content).hexdigest()

    (u_layer2_digest, (u_layer2_content, u_layer2_size)) = uploaded_layer2
    assert gzip.decompress(u_layer2_content) == test_tar_layer2_content
    assert u_layer2_size == len(u_layer2_content)
    assert u_layer2_digest == "sha256:" + hashlib.sha256(u_layer2_content).hexdigest()

    # check the uploaded manifest metadata and real content
    (uploaded_manifest,) = fake_registry.stored_manifests.items()
    (u_manifest_digest, (u_manifest_content, u_manifest_multiple)) = uploaded_manifest
    assert (
        u_manifest_digest
        == "sha256:" + hashlib.sha256(u_manifest_content.encode("utf8")).hexdigest()
    )
    assert u_manifest_multiple is False

    # the response from the function we're testing is the final remote digest
    assert main_call_result == u_manifest_digest

    u_manifest = json.loads(u_manifest_content)
    assert u_manifest["mediaType"] == MANIFEST_V2_MIMETYPE
    assert u_manifest["schemaVersion"] == 2

    assert u_manifest["config"] == {
        "digest": u_config_digest,
        "mediaType": CONFIG_MIMETYPE,
        "size": u_config_size,
    }

    assert u_manifest["layers"] == [
        {
            "digest": u_layer1_digest,
            "mediaType": LAYER_MIMETYPE,
            "size": u_layer1_size,
        },
        {
            "digest": u_layer2_digest,
            "mediaType": LAYER_MIMETYPE,
            "size": u_layer2_size,
        },
    ]

    # check the output logs
    emitter.assert_interactions(
        [
            call("progress", f"Getting the image from the local repo; size={image_size}"),
            call("progress_bar", "Reading image...", image_size),
            call("advance", image_read_from_dockerd_size_1),
            call("advance", image_read_from_dockerd_size_2),
            call("progress", "Extracting file 'config.yaml' from local tar (compress=False)"),
            call(
                "progress",
                f"Uploading config blob, size={u_config_size}, digest={u_config_digest}",
            ),
            call("progress", "Extracting file 'layer1.bin' from local tar (compress=True)"),
            call(
                "progress",
                f"Uploading layer blob 1/2, size={u_layer1_size}, digest={u_layer1_digest}",
            ),
            call("progress", "Extracting file 'layer2.bin' from local tar (compress=True)"),
            call(
                "progress",
                f"Uploading layer blob 2/2, size={u_layer2_size}, digest={u_layer2_digest}",
            ),
        ]
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_imagehandler_uploadfromlocal_no_config(emitter, tmp_path, monkeypatch):
    """Particular case of a manifest without config."""
    # fake an image in disk (a tar file with NO config, a layer, and a manifest)."""
    test_tar_image = tmp_path / "test-image.tar"
    test_tar_layer_content = b"fake layer content for the image"
    test_manifest_content = json.dumps(
        [
            {
                "Layers": ["layer.bin"],
            }
        ]
    ).encode("ascii")
    tar_file = tarfile.TarFile(test_tar_image, "w")
    tar_content = [
        ("manifest.json", test_manifest_content),
        ("layer.bin", test_tar_layer_content),
    ]
    for name, content in tar_content:
        ti = tarfile.TarInfo(name)
        ti.size = len(content)
        tar_file.addfile(ti, fileobj=io.BytesIO(content))
    tar_file.close()

    # return 200 with the image info
    image_size = test_tar_image.stat().st_size
    image_id = "test-image-id"
    image_info = {"Size": image_size, "Id": image_id, "foobar": "etc"}
    fakedockerd = FakeDockerd(image_id, image_info, test_tar_image.read_bytes())
    monkeypatch.setattr(registry, "LocalDockerdInterface", lambda: fakedockerd)

    fake_registry = FakeRegistry()
    im = ImageHandler(fake_registry)
    main_call_result = im.upload_from_local(image_info)

    # check the uploaded blob: just the compressed layer
    (uploaded_layer,) = fake_registry.stored_blobs.items()

    (u_layer_digest, (u_layer_content, u_layer_size)) = uploaded_layer
    assert gzip.decompress(u_layer_content) == test_tar_layer_content
    assert u_layer_size == len(u_layer_content)
    assert u_layer_digest == "sha256:" + hashlib.sha256(u_layer_content).hexdigest()

    # check the uploaded manifest metadata and real content
    (uploaded_manifest,) = fake_registry.stored_manifests.items()
    (u_manifest_digest, (u_manifest_content, u_manifest_multiple)) = uploaded_manifest
    assert (
        u_manifest_digest
        == "sha256:" + hashlib.sha256(u_manifest_content.encode("utf8")).hexdigest()
    )
    assert u_manifest_multiple is False

    # the response from the function we're testing is the final remote digest
    assert main_call_result == u_manifest_digest

    u_manifest = json.loads(u_manifest_content)
    assert u_manifest["mediaType"] == MANIFEST_V2_MIMETYPE
    assert u_manifest["schemaVersion"] == 2

    assert "config" not in u_manifest
    assert u_manifest["layers"] == [
        {
            "digest": u_layer_digest,
            "mediaType": LAYER_MIMETYPE,
            "size": u_layer_size,
        }
    ]

    # check the output logs
    emitter.assert_interactions(
        [
            call("progress", f"Getting the image from the local repo; size={image_size}"),
            call("progress_bar", "Reading image...", image_size),
            call("advance", image_size),
            call("progress", "Extracting file 'layer.bin' from local tar (compress=True)"),
            call(
                "progress",
                f"Uploading layer blob 1/1, size={u_layer_size}, digest={u_layer_digest}",
            ),
        ]
    )



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for common base commands functionality."""

import json

import craft_cli
from craft_cli import ArgumentParsingError, CraftError

JSON_FORMAT = "json"
FORMAT_HELP_STR = "Produce the result in the specified format (currently only 'json')"


class BaseCommand(craft_cli.BaseCommand):
    """Subclass this to create a new command.

    The following default attribute is provided beyond craft-cli ones:

    The subclass must be declared in the corresponding section of main.COMMAND_GROUPS.

    If the command may produce the result in a programmatic-friendly format, it
    should call the 'include_format_option' method to properly affect the parser and
    then emit only one message with the result of the 'format_content' method.
    """

    def format_content(self, fmt, content):
        """Format the content."""
        if fmt == JSON_FORMAT:
            return json.dumps(content, indent=4)
        raise ValueError("Specified format not supported.")

    def include_format_option(self, parser):
        """Add the 'format' option to this parser."""
        parser.add_argument(
            "--format",
            choices=[JSON_FORMAT],
            help=FORMAT_HELP_STR,
        )

    def _check_config(self, config_file: bool = False, bases: bool = False) -> None:
        """Check if valid config contents exists.

        - config_file: if True, check if a valid "charmcraft.yaml" file exists.
        - bases: if True, check if a valid "bases" in "charmcraft.yaml" exists.

        :raises ArgumentParsingError: if 'charmcraft.yaml' file is missing.
        :raises CraftError: if any specified config are missing or invalid.
        """
        if config_file and not self.config.project.config_provided:
            raise ArgumentParsingError(
                "The specified command needs a valid 'charmcraft.yaml' configuration file (in "
                "the current directory or where specified with --project-dir option); see "
                "the reference: https://discourse.charmhub.io/t/charmcraft-configuration/4138"
            )

        if bases and self.config.bases is None:
            raise CraftError(
                "The specified command needs a valid 'bases' in 'charmcraft.yaml' configuration "
                "file (in the current directory or where specified with --project-dir option)."
            )



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Pre-processing functions for charmcraft projects.

These functions are called from the Application class's `_extra_yaml_transform`
to do pre-processing on a charmcraft.yaml file before applying extensions.
"""
import pathlib
from typing import Any

from craft_application import errors, util

from charmcraft import const, utils


def add_default_parts(yaml_data: dict[str, Any]) -> None:
    """Apply the expected default parts to a project if it doesn't contain any.

    :param yaml_data: The raw YAML dictionary of the project.
    :returns: The same dictionary passed in, with necessary mutations.
    """
    if yaml_data.get("parts"):  # Only operate if there aren't any parts.
        return

    if yaml_data.get("type") == "bundle":
        yaml_data["parts"] = {"bundle": {"plugin": "bundle", "source": "."}}
    elif yaml_data.get("type") == "charm":
        # Only for backwards compatibility for bases charms.
        # Platforms charms expect parts to be explicit.
        if "bases" in yaml_data:
            yaml_data["parts"] = {"charm": {"plugin": "charm", "source": "."}}


def add_metadata(project_dir: pathlib.Path, yaml_data: dict[str, Any]) -> None:
    """Add the contents of metadata.yaml to a project data structure.

    :param project_dir: The directory containing charmcraft.yaml and possibly metadata.yaml.
    :param yaml_data: The current project data from charmcraft.yaml.
    """
    metadata_path = pathlib.Path(project_dir / "metadata.yaml")
    if not metadata_path.exists():
        return

    with metadata_path.open() as file:
        metadata_yaml = util.safe_yaml_load(file)
    if not isinstance(metadata_yaml, dict):
        raise errors.CraftValidationError(
            "Invalid file: 'metadata.yaml'",
            resolution="Ensure metadata.yaml meets the juju metadata.yaml specification.",
            docs_url="https://juju.is/docs/sdk/metadata-yaml",
            retcode=65,  # Data error, per sysexits.h
        )
    duplicate_fields = []
    for field in const.METADATA_YAML_MIGRATE_FIELDS:
        if field in yaml_data and field in metadata_yaml:
            duplicate_fields.append(field)
        yaml_data.setdefault(field, metadata_yaml.get(field))
    if duplicate_fields:
        raise errors.CraftValidationError(
            "Fields in charmcraft.yaml cannot be duplicated in metadata.yaml",
            details=f"Duplicate fields: {utils.humanize_list(duplicate_fields, 'and')}",
            resolution="Remove the duplicate fields from metadata.yaml.",
            retcode=65,  # Data error. per sysexits.h
        )


def add_bundle_snippet(project_dir: pathlib.Path, yaml_data: dict[str, Any]) -> None:
    """Add metadata from bundle.yaml to a bundle.

    :param yaml_data: The raw YAML dictionary of the project.
    :param project_dir: The Path to the directory containing charmcraft.yaml and bundle.yaml.
    :returns: The same dictionary passed in, with necessary mutations.
    """
    if yaml_data.get("type") != "bundle":
        return

    bundle_file = project_dir / const.BUNDLE_FILENAME
    if not bundle_file.is_file():
        raise errors.CraftValidationError(
            f"Missing 'bundle.yaml' file: {str(bundle_file)!r}",
            resolution="Create a 'bundle.yaml' file in the same directory as 'charmcraft.yaml'.",
            docs_url="https://juju.is/docs/sdk/create-a-charm-bundle",
            reportable=False,
            retcode=66,  # EX_NOINPUT from sysexits.h
        )

    with bundle_file.open() as bf:
        bundle = util.safe_yaml_load(bf)
    if not isinstance(bundle, dict):
        raise errors.CraftValidationError(
            "Incorrectly formatted 'bundle.yaml' file",
            resolution="Ensure 'bundle.yaml' matches the Juju 'bundle.yaml' format.",
            docs_url="https://juju.is/docs/sdk/charm-bundles",
            reportable=False,
            retcode=65,  # EX_DATAERR from sysexits.h
        )
    yaml_data["bundle"] = bundle


def add_config(project_dir: pathlib.Path, yaml_data: dict[str, Any]) -> None:
    """Add configuration options from config.yaml to existing YAML data.

    :param project_dir: The Path to the directory containing charmcraft.yaml
    :param yaml_data: The raw YAML dictionary of the project.
    :returns: The same dictionary passed in, with necessary mutations.
    """
    config_file = project_dir / const.JUJU_CONFIG_FILENAME
    if not config_file.exists():
        return

    if "config" in yaml_data:
        raise errors.CraftValidationError(
            f"Cannot specify 'config' section in 'charmcraft.yaml' when {const.JUJU_CONFIG_FILENAME!r} exists",
            resolution=f"Move all data from {const.JUJU_CONFIG_FILENAME!r} to the 'config' section in 'charmcraft.yaml'",
            docs_url="https://juju.is/docs/sdk/charmcraft-yaml",
            retcode=65,  # Data error, per sysexits.h
        )

    with config_file.open() as f:
        yaml_data["config"] = util.safe_yaml_load(f)


def add_actions(project_dir: pathlib.Path, yaml_data: dict[str, Any]) -> None:
    """Add actions from actions.yaml to existing YAML data.

    :param project_dir: The Path to the directory containing charmcraft.yaml
    :param yaml_data: The raw YAML dictionary of the project.
    :returns: The same dictionary passed in, with necessary mutations.
    """
    actions_file = project_dir / const.JUJU_ACTIONS_FILENAME
    if not actions_file.exists():
        return

    if "actions" in yaml_data:
        raise errors.CraftValidationError(
            f"Cannot specify 'actions' section in 'charmcraft.yaml' when {const.JUJU_ACTIONS_FILENAME!r} exists",
            resolution=f"Move all data from {const.JUJU_ACTIONS_FILENAME!r} to the 'actions' section in 'charmcraft.yaml'",
            docs_url="https://juju.is/docs/sdk/charmcraft-yaml",
            retcode=65,  # Data error, per sysexits.h
        )
    with actions_file.open() as f:
        actions = util.safe_yaml_load(f)
    if actions and isinstance(actions, dict):
        yaml_data["actions"] = actions



# Copyright 2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Provide utilities to measure performance in different parts of the app."""

import json
import uuid
from time import time
from typing import Any

# the base time of the started process: all the recorded times will be relative to
# this one (they will be easier to read and simple to understand the time passed since
# the process started until we started measuring)
_baseline = time()


class _Measurements:
    """Hold the measurements done and provide utilities around that structure."""

    def __init__(self):
        # ancestors list when a measure starts (last item is direct parent); the
        # first value is special, None, to reflect the "root", the rest are
        # measurement ids
        self.parents = [None]  # start with a unique "root"

        # simple dict to hold measurements information; the key is the measurement
        # id and each value holds all it info
        self.measurements = {}

    def start(self, msg: str, extra_info: dict[str, Any]):
        """Start a measurement."""
        this_id = uuid.uuid4().hex
        parent_id = self.parents[-1]
        self.parents.append(this_id)

        extra_info = {k: str(v) for k, v in extra_info.items()}
        self.measurements[this_id] = {
            "parent": parent_id,
            "msg": msg,
            "extra": extra_info,
            "tstart": time(),
            "tend": None,
        }
        return this_id

    def end(self, measurement_id):
        """Finish the indicated measurement."""
        if measurement_id != self.parents[-1]:
            raise ValueError("Overlapped measurements.")

        self.measurements[measurement_id]["tend"] = time()
        self.parents.pop()

    def dump(self, filename: str) -> None:
        """Dump the ongoing measurements to the specified file in a JSON format."""
        measurements = self.measurements.copy()
        for m in measurements.values():
            m["tstart"] -= _baseline
            m["tend"] -= _baseline
        measurements["__meta__"] = {"baseline": _baseline}
        with open(filename, "w") as fh:
            json.dump(measurements, fh, indent=4)

    def merge_from(self, filename: str) -> None:
        """Merge measurements from a file to the current ongoing structure."""
        with open(filename) as fh:
            to_merge = json.load(fh)

        # add the measurements from the file to this process ones, and while doing that
        # hook "root" measurements to current one and shift times
        current_id = self.parents[-1]
        sublayer_baseline = to_merge.pop("__meta__")["baseline"]

        for mid, data in to_merge.items():
            data["tstart"] += sublayer_baseline
            data["tend"] += sublayer_baseline
            if data["parent"] is None:
                data["parent"] = current_id
            self.measurements[mid] = data


# unique _Measurements object and external api
_measurements = _Measurements()
dump = _measurements.dump
merge_from = _measurements.merge_from


class Timer:
    """Get the timing of a block of code.

    It can work as a context manager that generates a measurement of the block of code inside
    its context, or as a decorator to get the timing of the whole decorated function.

    Usage as a context manager::

        with timer("some message", more="stuff", answer=42):
            ...

    Usage as a decorator::

        @timer("some message", foo="bar")
        def foo(...):
            ...

    In both cases the message is mandatory and the extra info optional.

    Also this class provides a `mark` method to sub-divide the context manager code
    block. How to use it::

        with timer("some message") as cm:
            ...
            cm.mark("first half done!")
            ...
    """

    def __init__(self, msg: str, **extra_info: dict[str, Any]):
        self.msg = msg
        self.extra_info = extra_info
        self.measurement_id = None

    def __enter__(self):
        self.measurement_id = _measurements.start(self.msg, self.extra_info)
        return self

    def __exit__(self, *exc):
        _measurements.end(self.measurement_id)

    def mark(self, msg: str, **extra_info: dict[str, Any]):
        """Mark middle measurements inside a contextual one."""
        # close the previous one, and start a new measure
        _measurements.end(self.measurement_id)
        self.measurement_id = _measurements.start(msg, extra_info)

    def __call__(self, func):
        """Decorate a function with self class to measure its execution."""

        def _f(*args, **kwargs):
            with self.__class__(self.msg, **self.extra_info):
                return func(*args, **kwargs)

        return _f



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Base command for Charmcraft commands."""
from __future__ import annotations

import craft_application.commands

from charmcraft import services


class CharmcraftCommand(craft_application.commands.AppCommand):
    """Base command for Charmcraft commands."""

    format_option: bool = False
    _services: services.CharmcraftServiceFactory

    def fill_parser(self, parser) -> None:
        """Fill the argument parser for this command."""
        super().fill_parser(parser)
        if self.format_option:
            parser.add_argument(
                "--format",
                choices=["json"],
                help="Produce the result in the specified format (currently only 'json')",
            )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import platform
import sys
from textwrap import dedent
from typing import Any

import pytest
from overrides import override

from charmcraft import extensions
from tests.extensions.test_extensions import FakeExtension


class TestExtension(FakeExtension):
    """A fake test Extension that has complete behavior"""

    name = "test-extension"
    bases = [("ubuntu", "22.04")]

    @override
    def get_root_snippet(self) -> dict[str, Any]:
        """Return the root snippet to apply."""
        return {"terms": ["https://example.com/test"]}


@pytest.fixture
def fake_extensions(stub_extensions):
    extensions.register(TestExtension.name, TestExtension)


@pytest.mark.xfail(
    platform.system() == "Windows" and sys.version_info < (3, 11),
    reason="'os' module doesn't have EX_OK on Windows until 3.11",
)
@pytest.mark.parametrize(
    ("charmcraft_yaml", "expected"),
    [
        (
            dedent(
                f"""
                name: test-charm-name
                type: charm
                summary: test-summary
                description: test-description
                extensions: [{TestExtension.name}]
                base: ubuntu@22.04
                platforms:
                  amd64:
                """
            ),
            dedent(
                """\
                name: test-charm-name
                summary: test-summary
                description: test-description
                base: ubuntu@22.04
                platforms:
                  amd64:
                    build-on:
                    - amd64
                    build-for:
                    - amd64
                parts: {}
                type: charm
                terms:
                - https://example.com/test
                """
            ),
        )
    ],
)
def test_expand_extensions_simple(
    monkeypatch, new_path, app, fake_extensions, emitter, charmcraft_yaml, expected
):
    """Expand a charmcraft.yaml with a single extension."""
    (new_path / "charmcraft.yaml").write_text(charmcraft_yaml)
    monkeypatch.setattr(sys, "argv", ["charmcraft", "expand-extensions"])

    app.run()

    emitter.assert_message(expected)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Service class for handling OCI images."""

import dataclasses
import io
import logging
import subprocess
from collections.abc import Sequence

import craft_application
import docker  # type: ignore[import-untyped]
import docker.errors  # type: ignore[import-untyped]
import docker.models.images  # type: ignore[import-untyped]
from overrides import override

from charmcraft import const, errors, utils

logger = logging.getLogger("charmcraft")


@dataclasses.dataclass(frozen=True)
class OCIMetadata:
    """Metadata about an OCI image.

    :param path: A skopeo-compatible image path
    :param digest: A digest string for the image
    :param architectures: A list of architectures for Linux platforms of the image.
    """

    path: str
    digest: str
    architectures: Sequence[const.CharmArch]


class ImageService(craft_application.AppService):
    """Service to handle OCI images.

    This service is mostly a craft-application friendly wrapper around skopeo.
    Use of this service requires the app to have access to skopeo.
    """

    _skopeo: utils.Skopeo
    _docker: docker.DockerClient

    @override
    def setup(self) -> None:
        """Set up the image service."""
        super().setup()
        self._skopeo = utils.Skopeo(insecure_policy=True)
        self._docker = docker.from_env()

    def copy(
        self,
        source_image: str,
        destination_image: str,
        stdout: io.FileIO,
        *,
        dest_username: str | None = None,
        dest_password: str | None = None,
    ):
        """Use Skopeo to copy an image.

        :param source_image: A skopeo-accepted source image string.
        :param destination_image: A skopeo-accepted destination image string.
        :param stdout: A stream to use as skopeo's stdout

        This is designed to be used with craft-cli's emit.open_stream or similar.
        For example:
        >>> with emit.open_stream("Uploading image") as stream:
        ...     image_service.copy(source, dest, stdout=stream)
        However, it can be used in other ways by passing any file-like object as the stream.
        """
        self._skopeo.copy(
            source_image=source_image,
            destination_image=destination_image,
            stdout=stdout,
            stderr=subprocess.PIPE,
            dest_username=dest_username,
            dest_password=dest_password,
            preserve_digests=True,
        )

    def get_maybe_id_from_docker(self, name: str) -> str | None:
        """Get the ID of an image from Docker.

        :param name: Any string Docker recognises as the image name.
        :returns: An image digest or None

        The digest will match the OCI digest spec:
        https://github.com/opencontainers/image-spec/blob/main/descriptor.md#digests
        """
        try:
            image = self._docker.images.get(name)
        except docker.errors.ImageNotFound:
            return None
        return image.id

    def inspect(self, image: str) -> OCIMetadata:
        """Inspect an image with Skopeo and return the relevant metadata.

        :param image: A skopeo-friendly image name
        :returns: An OCIMetadata object containing metadata about the image.
        """
        try:
            raw_manifest = self._skopeo.inspect(image, raw=True)
            image_info = self._skopeo.inspect(image)
        except errors.SubprocessError as exc:
            raise errors.CraftError(
                "Could not inspect OCI image.", details=f"{exc.message}\n{exc.details}"
            )
        try:
            image_digest = image_info["Digest"]
        except KeyError:
            raise errors.CraftError("Could not get digest for image.")
        if "manifests" in raw_manifest:  # Multi-arch OCI image
            architectures: list[const.CharmArch] = []
            for child in raw_manifest["manifests"]:
                platform = child.get("platform", {})
                if platform.get("os") != "linux":
                    continue
                architectures.append(const.CharmArch(platform["architecture"]))
            if not architectures:
                raise errors.CraftError("No architectures found in image for Linux OS.")
        else:
            architectures = [const.CharmArch(image_info["Architecture"])]

        return OCIMetadata(
            path=image,
            digest=image_digest,
            architectures=architectures,
        )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for extension commands."""
import argparse
import textwrap

import pytest

from charmcraft import application, extensions
from charmcraft.application import commands


def create_extension(ext_name, bases, experimental):
    class FakeExtension(extensions.Extension):
        name = ext_name

        @staticmethod
        def get_supported_bases() -> list[tuple[str, str]]:
            return bases + experimental

        @staticmethod
        def is_experimental(base: tuple[str, str] | None) -> bool:
            return base in experimental

    return FakeExtension


@pytest.fixture(autouse=True, scope="module")
def registered_extensions():
    default_extensions = {
        name: extensions.get_extension_class(name) for name in extensions.get_extension_names()
    }
    for ext in default_extensions:
        extensions.unregister(ext)
    fake_extensions = [
        create_extension("f1", [("ubuntu", "22.04")], [("ubuntu", "24.04")]),
        create_extension("f2", [], [("almalinux", "9")]),
    ]
    for ext in fake_extensions:
        extensions.register(ext.name, ext)
    yield fake_extensions
    for ext in fake_extensions:
        extensions.unregister(ext.name)
    for name, cls in default_extensions.items():
        extensions.register(name, cls)


@pytest.mark.parametrize(
    ("fmt", "expected"),
    [
        (
            None,
            textwrap.dedent(
                """\
            Extension name    Supported bases    Experimental bases
            ----------------  -----------------  --------------------
            f1                ubuntu@22.04       ubuntu@24.04
            f2                                   almalinux@9"""
            ),
        ),
        (
            "json",
            textwrap.dedent(
                """\
            [
                {
                    "name": "f1",
                    "bases": [
                        "ubuntu@22.04"
                    ],
                    "experimental_bases": [
                        "ubuntu@24.04"
                    ]
                },
                {
                    "name": "f2",
                    "bases": [],
                    "experimental_bases": [
                        "almalinux@9"
                    ]
                }
            ]"""
            ),
        ),
    ],
)
def test_list_extensions(emitter, fmt, expected):
    cmd = commands.ListExtensionsCommand(
        {
            "app": application.APP_METADATA,
            "services": None,
        }
    )

    cmd.run(argparse.Namespace(format=fmt))

    emitter.assert_message(expected)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Service class for packing."""
from __future__ import annotations

import pathlib
import tempfile
import zipfile
from collections.abc import Container, Iterator

import craft_application

from charmcraft import errors, linters, models
from charmcraft.models.lint import CheckResult


class AnalysisService(craft_application.AppService):
    """Business logic for creating packages."""

    _project: models.CharmcraftProject  # type: ignore[assignment]

    def __init__(  # (too many arguments)
        self, app: craft_application.AppMetadata, services: craft_application.ServiceFactory
    ) -> None:
        super().__init__(app, services)

    def lint_directory(
        self, path: pathlib.Path, *, ignore: Container[str] = (), include_ignored: bool = True
    ) -> Iterator[CheckResult]:
        """Lint an unpacked charm in the given directory."""
        for checker, run in self._gen_checkers(ignore=ignore):
            if run:
                yield checker.get_result(path)
            elif include_ignored:
                yield checker.get_ignore_result()

    def lint_file(
        self, path: pathlib.Path, *, ignore: Container[str] = (), include_ignored: bool = True
    ) -> Iterator[CheckResult]:
        """Lint a packed charm.

        :param path: The path to the file
        :param ignore: a list of checker names to ignore.
        :param include_ignored: Whether to include ignored values in the output

        raises: FileNotFoundError if the file doesn't exist
        """
        path = path.resolve(strict=True)

        with tempfile.TemporaryDirectory(prefix=f"charmcraft_{path.name}_") as directory:
            directory_path = pathlib.Path(directory)
            try:
                with zipfile.ZipFile(path) as zip_file:
                    zip_file.extractall(directory_path)
                    # fix permissions as extractall does not keep them (see https://bugs.python.org/issue15795)
                    for name in zip_file.namelist():
                        info = zip_file.getinfo(name)
                        inside_zip_mode = info.external_attr >> 16
                        extracted_file = directory_path / name
                        current_mode = extracted_file.stat().st_mode
                        if current_mode != inside_zip_mode:
                            extracted_file.chmod(inside_zip_mode)
            except zipfile.BadZipfile as exc:
                raise errors.CraftError(
                    f"Cannot open charm file '{path}': {exc.args[0]}",
                    resolution=f"Check the charm file at {path}",
                    reportable=False,
                )
            yield from self.lint_directory(
                directory_path, ignore=ignore, include_ignored=include_ignored
            )

    @staticmethod
    def _gen_checkers(ignore: Container[str]) -> Iterator[tuple[linters.BaseChecker, bool]]:
        """Generate the checker classes to run, in their correct order."""
        for cls in linters.CHECKERS:
            run_linter = cls.name not in ignore
            yield cls(), run_linter



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for packing."""
import sys
import zipfile

import pytest
import yaml
from craft_application import util

from charmcraft import models, utils

CURRENT_PLATFORM = utils.get_os_platform()


@pytest.mark.xfail(
    sys.platform != "linux", reason="https://github.com/canonical/charmcraft/issues/1552"
)
@pytest.mark.parametrize(
    ("bundle_yaml", "filename"),
    [
        ("{}", "bundle.zip"),
        ("name: my-bundle", "my-bundle.zip"),
    ],
)
def test_build_basic_bundle(monkeypatch, capsys, app, new_path, bundle_yaml, filename):
    (new_path / "charmcraft.yaml").write_text("type: bundle")
    (new_path / "bundle.yaml").write_text(bundle_yaml)

    monkeypatch.setenv("CRAFT_DEBUG", "1")
    monkeypatch.setattr("sys.argv", ["charmcraft", "pack", "--verbosity=trace"])

    app.configure({})
    if app.run() != 0:
        raise ValueError(capsys.readouterr())

    with zipfile.ZipFile(filename) as bundle_zip:
        actual_bundle_yaml = bundle_zip.read("bundle.yaml").decode()

    assert actual_bundle_yaml == bundle_yaml


@pytest.mark.parametrize(
    ("charmcraft_project", "platform"),
    [
        pytest.param(
            {
                "type": "charm",
                "name": "my-charm",
                "summary": "A test charm",
                "description": "A charm for testing",
                "bases": [
                    {
                        "build-on": [{"name": "ubuntu", "channel": "22.04"}],
                        "run-on": [
                            {"name": "ubuntu", "channel": "22.04", "architectures": ["amd64"]}
                        ],
                    }
                ],
            },
            "ubuntu-22.04-amd64",
            marks=pytest.mark.skipif(
                CURRENT_PLATFORM.release != "22.04", reason="Bases charm only tested on jammy."
            ),
            id="bases-charm",
        ),
        pytest.param(
            {
                "type": "charm",
                "name": "my-charm",
                "summary": "A test charm",
                "description": "A charm for testing",
                "base": "ubuntu@22.04",
                "platforms": {
                    "ubuntu-22.04-amd64": {"build-on": ["amd64"], "build-for": ["amd64"]}
                },
                "parts": {},
            },
            "ubuntu-22.04-amd64",
            marks=pytest.mark.skipif(
                CURRENT_PLATFORM.release != "22.04", reason="Jammy charms only tested on jammy"
            ),
            id="platforms-jammy-charm",
        ),
        pytest.param(
            {
                "type": "charm",
                "name": "my-charm",
                "summary": "A test charm",
                "description": "A charm for testing",
                "base": "ubuntu@22.04",
                "platforms": {util.get_host_architecture(): None},
                "parts": {},
            },
            util.get_host_architecture(),
            marks=pytest.mark.skipif(
                CURRENT_PLATFORM.release != "22.04", reason="Jammy charms only tested on jammy"
            ),
            id="platforms-jammy-basic",
        ),
        pytest.param(
            {
                "type": "charm",
                "name": "my-charm",
                "summary": "A test charm",
                "description": "A charm for testing",
                "base": "ubuntu@24.04",
                "build-base": "ubuntu@devel",
                "platforms": {util.get_host_architecture(): None},
                "parts": {},
            },
            util.get_host_architecture(),
            marks=pytest.mark.skipif(
                CURRENT_PLATFORM.release != "24.04", reason="Noble charm needs noble"
            ),
            id="platforms-noble",
        ),
    ],
)
@pytest.mark.skipif(
    CURRENT_PLATFORM.system != "ubuntu",
    reason="Basic charm tests use destructive mode.",
)
def test_build_basic_charm(
    monkeypatch, emitter, new_path, charmcraft_project, service_factory, app, platform
):
    (new_path / "charmcraft.yaml").write_text(yaml.dump(charmcraft_project))
    service_factory.project = models.CharmcraftProject.unmarshal(charmcraft_project)

    monkeypatch.setenv("CRAFT_DEBUG", "1")
    monkeypatch.setattr(
        "sys.argv", ["charmcraft", "pack", "--destructive-mode", f"--platform={platform}"]
    )

    app.configure({})
    assert app.run() == 0

    with zipfile.ZipFile(new_path / f"my-charm_{platform}.charm") as charm_zip:
        metadata = yaml.safe_load(charm_zip.read("metadata.yaml"))
        manifest = yaml.safe_load(charm_zip.read("manifest.yaml"))

    emitter.assert_progress(f"Packing charm my-charm_{platform}.charm")

    assert "bases" in manifest
    if "platforms" in charmcraft_project:
        base_name = manifest["bases"][0]["name"]
        base_version = manifest["bases"][0]["channel"]
        assert f"{base_name}@{base_version}" == charmcraft_project["base"]
    else:
        assert manifest["bases"][0]["name"] in platform
        assert manifest["bases"][0]["channel"] in platform
    assert manifest["bases"][0]["architectures"][0] in platform

    assert metadata["name"] == charmcraft_project["name"]
    assert metadata["summary"] == charmcraft_project["summary"]
    assert metadata["description"] == charmcraft_project["description"]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Store module for Charmcraft."""

from charmcraft.store.client import build_user_agent, AnonymousClient, Client
from charmcraft.store import models
from charmcraft.store.models import LibraryMetadataRequest
from charmcraft.store.registry import (
    OCIRegistry,
    HashingTemporaryFile,
    LocalDockerdInterface,
    ImageHandler,
)
from charmcraft.store.store import Store, AUTH_DEFAULT_TTL, AUTH_DEFAULT_PERMISSIONS

__all__ = [
    "build_user_agent",
    "AnonymousClient",
    "Client",
    "OCIRegistry",
    "HashingTemporaryFile",
    "ImageHandler",
    "LocalDockerdInterface",
    "AUTH_DEFAULT_PERMISSIONS",
    "AUTH_DEFAULT_TTL",
    "Store",
    "models",
    "LibraryMetadataRequest",
]



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Service classes charmcraft."""
from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING
from charmcraft.services.remotebuild import RemoteBuildService

from craft_application import ServiceFactory

from .analysis import AnalysisService
from .image import ImageService
from .lifecycle import LifecycleService
from .package import PackageService
from .provider import ProviderService
from .store import StoreService
from .. import models


@dataclasses.dataclass
class CharmcraftServiceFactory(ServiceFactory):
    """Factory class for lazy-loading Charmcraft services."""

    PackageClass: type[PackageService] = PackageService
    LifecycleClass: type[LifecycleService] = LifecycleService
    ProviderClass: type[ProviderService] = ProviderService
    AnalysisClass: type[AnalysisService] = AnalysisService
    StoreClass: type[StoreService] = StoreService
    RemoteBuildClass: type[RemoteBuildService] = RemoteBuildService
    ImageClass: type[ImageService] = ImageService

    if TYPE_CHECKING:
        # Cheeky hack that lets static type checkers report the correct types.
        # Any apps that add their own services should do this too.
        analysis: AnalysisService = None  # type: ignore[assignment]
        image: ImageService = None  # type: ignore[assignment]
        lifecycle: LifecycleService = None  # type: ignore[assignment]
        package: PackageService = None  # type: ignore[assignment]
        project: models.CharmcraftProject = None  # type: ignore[assignment]
        provider: ProviderService = None  # type: ignore[assignment]
        store: StoreService = None  # type: ignore[assignment]


__all__ = [
    "AnalysisService",
    "ImageService",
    "LifecycleService",
    "PackageService",
    "ProviderService",
    "CharmcraftServiceFactory",
]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for CLI-related utilities."""
import datetime
import json
from unittest.mock import call, patch

import dateutil.parser
import pytest
import tabulate
from hypothesis import given, strategies

from charmcraft.utils.cli import (
    ChoicesList,
    OutputFormat,
    ResourceOption,
    SingleOptionEnsurer,
    confirm_with_user,
    format_content,
    format_timestamp,
    humanize_list,
)

OUTPUT_VALUES_STRATEGY = strategies.one_of(
    strategies.none(), strategies.integers(), strategies.text()
)
BASIC_LISTS_STRATEGY = strategies.lists(OUTPUT_VALUES_STRATEGY)
BASIC_DICTS_STRATEGY = strategies.dictionaries(
    strategies.one_of(OUTPUT_VALUES_STRATEGY), strategies.one_of(OUTPUT_VALUES_STRATEGY)
)
COMPOUND_DICTS_STRATEGY = strategies.dictionaries(
    strategies.one_of(OUTPUT_VALUES_STRATEGY),
    strategies.one_of(OUTPUT_VALUES_STRATEGY, BASIC_DICTS_STRATEGY),
)


@pytest.fixture
def mock_isatty():
    with patch("sys.stdin.isatty", return_value=True) as mock_isatty:
        yield mock_isatty


@pytest.fixture
def mock_input():
    with patch("charmcraft.utils.cli.input", return_value="") as mock_input:
        yield mock_input


@pytest.fixture
def mock_is_charmcraft_running_in_managed_mode():
    with patch(
        "charmcraft.utils.cli.is_charmcraft_running_in_managed_mode", return_value=False
    ) as mock_managed:
        yield mock_managed


def test_singleoptionensurer_convert_ok():
    """Work fine with one call, convert as expected."""
    soe = SingleOptionEnsurer(int)
    assert soe("33") == 33


def test_singleoptionensurer_too_many():
    """Raise an error after one ok call."""
    soe = SingleOptionEnsurer(int)
    assert soe("33") == 33
    with pytest.raises(ValueError) as cm:
        soe("33")
    assert str(cm.value) == "the option can be specified only once"


def test_resourceoption_convert_ok():
    """Convert as expected."""
    r = ResourceOption()("foo:13")
    assert r.name == "foo"
    assert r.revision == 13


@pytest.mark.parametrize(
    "value",
    [
        pytest.param("foo15", id="no separation"),
        pytest.param("foo:", id="no revision"),
        pytest.param("foo:x3", id="non-integer revision"),
        pytest.param("foo:-1", id="negative revisions are not allowed"),
        pytest.param(":15", id="no name"),
        pytest.param("  :15", id="no name, really!"),
        pytest.param("foo:bar:15", id="invalid name"),
    ],
)
def test_resourceoption_convert_error(value):
    """Error while converting."""
    with pytest.raises(ValueError) as cm:
        ResourceOption()(value)
    assert str(cm.value) == (
        "the resource format must be <name>:<revision> (revision being a non-negative integer)"
    )


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("abc", ["abc"]),
        ("abc,def", ["abc", "def"]),
    ],
)
def test_choices_list_success(value, expected):
    choices_list = ChoicesList(["abc", "def", "ghi"])

    assert choices_list(value) == expected


def test_choices_list_invalid_values():
    choices_list = ChoicesList({})

    with pytest.raises(ValueError, match="^invalid values: abc$"):
        choices_list("abc")


def test_confirm_with_user_defaults_with_tty(mock_input, mock_isatty):
    mock_input.return_value = ""
    mock_isatty.return_value = True

    assert confirm_with_user("prompt", default=True) is True
    assert mock_input.mock_calls == [call("prompt [Y/n]: ")]
    mock_input.reset_mock()

    assert confirm_with_user("prompt", default=False) is False
    assert mock_input.mock_calls == [call("prompt [y/N]: ")]


def test_confirm_with_user_defaults_without_tty(mock_input, mock_isatty):
    mock_isatty.return_value = False

    assert confirm_with_user("prompt", default=True) is True
    assert confirm_with_user("prompt", default=False) is False

    assert mock_input.mock_calls == []


@pytest.mark.parametrize(
    ("user_input", "expected"),
    [
        ("y", True),
        ("Y", True),
        ("yes", True),
        ("YES", True),
        ("n", False),
        ("N", False),
        ("no", False),
        ("NO", False),
    ],
)
def test_confirm_with_user(user_input, expected, mock_input, mock_isatty):
    mock_input.return_value = user_input

    assert confirm_with_user("prompt") == expected
    assert mock_input.mock_calls == [call("prompt [y/N]: ")]


def test_confirm_with_user_errors_in_managed_mode(mock_is_charmcraft_running_in_managed_mode):
    mock_is_charmcraft_running_in_managed_mode.return_value = True

    with pytest.raises(RuntimeError):
        confirm_with_user("prompt")


def test_confirm_with_user_pause_emitter(mock_isatty, emitter):
    """The emitter should be paused when using the terminal."""
    mock_isatty.return_value = True

    def fake_input(prompt):
        """Check if the Emitter is paused."""
        assert emitter.paused
        return ""

    with patch("charmcraft.utils.cli.input", fake_input):
        confirm_with_user("prompt")


def test_timestampstr_simple():
    """Converts a timestamp without timezone."""
    source = datetime.datetime(2020, 7, 3, 20, 30, 40)
    result = format_timestamp(source)
    assert result == "2020-07-03T20:30:40Z"


def test_timestampstr_utc():
    """Converts a timestamp with UTC timezone."""
    source = dateutil.parser.parse("2020-07-03T20:30:40Z")
    result = format_timestamp(source)
    assert result == "2020-07-03T20:30:40Z"


def test_timestampstr_nonutc():
    """Converts a timestamp with other timezone."""
    source = dateutil.parser.parse("2020-07-03T20:30:40+03:00")
    result = format_timestamp(source)
    assert result == "2020-07-03T17:30:40Z"


@pytest.mark.parametrize(
    ("items", "conjunction", "expected"),
    (
        (["foo"], "xor", "'foo'"),
        (["foo", "bar"], "xor", "'bar' xor 'foo'"),
        (["foo", "bar", "baz"], "xor", "'bar', 'baz' xor 'foo'"),
        (["foo", "bar", "baz", "qux"], "xor", "'bar', 'baz', 'foo' xor 'qux'"),
    ),
)
def test_humanize_list_ok(items, conjunction, expected):
    """Several successful cases."""
    assert humanize_list(items, conjunction) == expected


def test_humanize_list_empty():
    """Calling to humanize an empty list is an error that should be explicit."""
    with pytest.raises(ValueError):
        humanize_list([], "whatever")


@given(
    strategies.one_of(
        strategies.none(),
        strategies.integers(),
        strategies.text(),
        strategies.dates(),
    )
)
def test_format_content_string(content):
    assert format_content(content, None) == str(content)


@given(
    strategies.one_of(
        strategies.none(),
        strategies.integers(),
        strategies.floats(),
        strategies.text(),
        BASIC_LISTS_STRATEGY,
        COMPOUND_DICTS_STRATEGY,
    )
)
def test_format_content_json(content):
    assert format_content(content, "json") == json.dumps(content, indent=4)


@given(
    strategies.lists(
        strategies.fixed_dictionaries(
            {
                "first column": OUTPUT_VALUES_STRATEGY,
                "Column 2": OUTPUT_VALUES_STRATEGY,
                3: OUTPUT_VALUES_STRATEGY,
                None: OUTPUT_VALUES_STRATEGY,
            }
        )
    )
)
def test_format_content_table(content):
    assert format_content(content, OutputFormat.TABLE) == tabulate.tabulate(
        content, headers="keys"
    )


@pytest.mark.parametrize("fmt", ["yolo", 0])
def test_format_content_invalid(fmt):
    with pytest.raises(ValueError, match="^Unknown output format "):
        format_content(None, fmt)
